-- Habitat Wars
-- By Tim Sandford, 2019
-- Comments to timsandford42@hotmail.com
function love.run()  -- Main function
  gVERSION, gPATCH, gREVISION, gCODENAME = love.getVersion()
  if gVERSION >=  11 then  -- Colour fix for 11.0 onwards
    local oldSetColor = love.graphics.setColor
    love.graphics.setColor = function (r, g, b, a)     
      if type(r)=="table" then
        r = r[1] / 255
        g = r[2] / 255
        b = r[3] / 255
        a = (r[4] or 255) / 255
      else
        r = r / 255
        g = g / 255
        b = b / 255
        a = (a or 255) / 255
      end
      oldSetColor(r,g,b,a)
    end
  end
  math.randomseed(os.time())  -- True randomness
  love.load(love.arg.parseGameArguments(arg), arg) -- load variables etc.
  while true do  -- Main loop
    if love.event then -- Process events
      love.event.pump()
      for name, a,b,c,d,e,f in love.event.poll() do
        if name == "quit" then
          if not love.quit or not love.quit() then
            return a
          end
        end
        love.handlers[name](a,b,c,d,e,f)
      end
    end
    love.update() -- Update
    love.graphics.origin() -- Not sure, probably reset origin?
    love.graphics.clear(love.graphics.getBackgroundColor()) -- Clears the screen
    love.draw() -- Collect draw commands
    love.graphics.present() -- Present the graphics on the screen!
    love.timer.sleep(0.001) -- This is to lower the load on the CPU, but I could edit it :)	
  end
end

function love.load(arg)  -- Load everything in
  local i, s
  io.stdout:setvbuf("no")                                         -- Print messages immediately, allows tracing bugs
  if arg[#arg] == "-debug" then require("mobdebug").start() end   -- Zerobrane debugging
  gGAMEMODE = "Title"                                             -- Set gamemode
  gSCREENWIDTH = love.graphics.getWidth()                         -- Get screensize
  gSCREENHEIGHT = love.graphics.getHeight()                       -- Get screensize
  gRATIOX = gSCREENWIDTH / 800                                    -- Scale graphics to screen
  gRATIOY = gSCREENHEIGHT / 600                                   -- Scale graphics to screen
  gFADESPEED = 10                                                 -- Speed of fade in effect
  gFPS = 45                                                       -- Frames per second
  gTIMER = love.timer.getTime()                                   -- Use timer in love.update for speed of game
  gSOUNDON = true                                                 -- Sounds on
  gMUSICON = true                                                 -- Music on
  gFLASH = 255			                                               -- Flashing text transparency
  gFLASH_z = 10 			                                             -- Speed of change for flashing text
  gFLASH_min = 80                                                 -- Minimum transparency of anything flashing
  gFLASH_max = 255                                                -- Maximum transparency of anything flashing
  gCARDWIDTH = 80                                                 -- Width of card standard zoom
  gCARDHEIGHT = 112                                               -- Height of card standard zoom
  gHANDWIDTH = 260                                                -- Width of space for hand
  gHANDHEIGHT = 170                                               -- Height of space for hand
  gHANDSLANT = 15                                                 -- Curve of cards across hand
  gMAXCARDSINHAND = 7                                             -- Maximum cards a player can hold in their hand
  gMAXCARDSINCOMBAT = 6                                           -- Maximum number of allies
  gCOMBATWIDTH = 400                                              -- Width of space for hand
  gANIMATION_TIME = math.floor(0.3 * gFPS)                        -- How long an animation takes in seconds
  gSHUFFLE = 2000			                                            -- Number of swaps in the deck
  gCOINS = 0                                                      -- Amount of coins the player can spend
  gDELAYACTION = 0                                                -- Counter for delayed movements
  gDECKSIZE = 39                                                  -- 40 including the hero
  gPOPUP_COUNT = 0                                                -- Countdown until msg disappears
  gPOPUP_TIME = 1.7*gFPS                                          -- Timer for popup messages
  gFLOAT_TIMER = 100                                              -- How long floaty text stays around for
  AImax_loops = 200                                               -- A.I. level, set by difficulty later
  
  tHABITAT = {}
  tCARDS = {}
  tSAVES = {}
  tEFFECTS = {"Inactive", "Hidden", "Limp", "Fear", "Well Fed", "Light", "Blinded", "Defender", "Nighttime", "Daytime", "Venom", "Venom", "Venom", "Rage"}
  tRESULTS = {"Cannot initiate attack", "Cannot be targeted", "Cannot attack or defend", "-50% attack", "+25% attack", "+33% attack", "-33% attack", "+2 power/health", "+50% attack", "-50% attack", "-1HP per turn", "-3HP per turn", "-5HP per turn", "+25% attack"}
  
  xHAND = {400, 400}
  yHAND = {500, 80}
  xDECK = {60, 60}
  yDECK = {520, 90}
  xGRAVE = {730, 730}
  yGRAVE = {520, 80}
  xCOMBAT = {400, 400}
  yCOMBAT = {300+(gCARDHEIGHT*0.7), 300-(gCARDHEIGHT*0.7)}
  xHERO = {730, 730}
  yHERO = {300+(gCARDHEIGHT*0.7), 300-(gCARDHEIGHT*0.7)}
  
  iCARDBACK = love.graphics.newImage("gui/CardBack.png")
  iUNKNOWN = love.graphics.newImage("gui/QuestionMark.png")
  iBLANK = love.graphics.newImage("gui/Blank.png")
  iSQUARE = love.graphics.newImage("gui/CardSquare.png")
  iTILE = love.graphics.newImage("gui/BlankTile.png")
  
  readcards()
  resetmainmenu()
  readhabitats()
  loadlevels()
  refreshsavefiles()
  loadcards()
  
  if next(tCARDS) ~= nil then
    for i = 1, #tCARDS, 1 do
      blankcard(tCARDS[i], true)
    end
  end
  
  sSOUND = {}                                                     -- Load sounds
  for i = 1, 99, 1 do
    if i < 10 then s = "sounds/sound0"..i..".ogg"
    else s = "sounds/sound"..i..".ogg" end
    if love.filesystem.getInfo(s, "file") then sSOUND[i] = love.audio.newSource(s, "static") end
  end
  
  if love.filesystem.getInfo("gui/cursor.png", "file") then       -- Mouse pointer
    gCURSOR = love.mouse.newCursor("gui/cursor.png", 0, 0)
    love.mouse.setCursor(gCURSOR)
  end
  love.mouse.setVisible(true)
  
  fTINY = love.graphics.newFont("fonts/font.ttf", 9)              -- tiny font
  fSMALL = love.graphics.newFont("fonts/font.ttf", 12)            -- small font
  fDEFAULT = love.graphics.newFont("fonts/font.ttf", 16)          -- default font
  fHEADING = love.graphics.newFont("fonts/font.ttf", 28)          -- heading font
  fTITLE = love.graphics.newFont("fonts/font.ttf", 40)            -- title font
  
  iICON = {}                                                      -- icon images
  for i = 1, 99, 1 do
    if i < 10 then s = "gui/icon0"..i..".png"
    else s = "gui/icon"..i..".png" end
    if love.filesystem.getInfo(s, "file") then iICON[i] = love.graphics.newImage(s) end
  end
end

function resetmainmenu()  -- Reset variables for main menu / new game
  gFADE = 255                                                     -- Fade in effect
  gTYPING = false                                                 -- Are they typing in text?
  gTYPEDTEXT = ""                                                 -- What are they typing?
  gDRAGGING = false                                               -- Are they dragging a card?
  gHOLDING = 0                                                    -- How long have they been dragging a card?
  gZOOMED = false                                                 -- Are they zoomed in on a card?
  gACTION = ""                                                    -- What action has the player/AI done?
  bACTION = ""                                                    -- Backup for gACTION in checkpoint()
  hACTION = ""                                                    -- Backup for gACTION in harfway()
  gSACRIFICED = false                                             -- Has the player sacrificed a card yet?
  bSACRIFICED = false                                             -- Backup for sacrificed in checkpoint()
  hSACRIFICED = false                                             -- Backup for sacrificed in harfway()
  gCEASEFIRE = 0                                                  -- Prevents attacking
  bCEASEFIRE = 0                                                  -- Backup for ceasefire in checkpoint()
  hCEASEFIRE = 0                                                  -- Backup for ceasefire in harfway()
  gLEVEL_CURSOR = 0                                               -- Select a level to play
  gDECK_SELECT = 0                                                -- Select a deck (habitat)
  gDIFFICULTY = 1                                                 -- 1=Easy, 2=Medium, 3=Hard
  gFILEMODE = ""                                                  -- Are they loading, saving or erasing?
  gPOPUP_MSG = ""                                                 -- Display a message e.g. error
  gTURN =  math.random(2)                                         -- Whose turn is it
  gENDING = 0                                                     -- Countdown at end of game
  gDAYNIGHT = 0                                                   -- Day 6-10, Night 1-5
  tCONTROL = {"Player", "AI"}                                     -- Singleplayer / Pass n Play
  tUNLOCKED = {}
  tCLICK = {}
  tLIST = {}
  tDECK = {{}, {}}
  tHAND = {{}, {}}
  tGRAVE = {{}, {}}
  tCOMBAT= {{}, {}}
  tHERO = {{}, {}}
  tBUDGET = {0, 0}
  tPOINTS = {0, 0}
  bDECK = {{}, {}}
  bHAND = {{}, {}}
  bGRAVE = {{}, {}}
  bCOMBAT = {{}, {}}
  bHERO = {{}, {}}
  bBUDGET = {}
  bPOINTS = {}
  hDECK = {{}, {}}
  hHAND = {{}, {}}
  hGRAVE = {{}, {}}
  hCOMBAT = {{}, {}}
  hHERO = {{}, {}}
  hBUDGET = {}
  hPOINTS = {}
  AIsimulation = false
  AIbest_score = 0
  AIbest_start = {}
  AIloop = 0
  AIlastmove = love.timer.getTime()
  loaddeck("Autosave", 1)
end

function readcards()     -- Read card templates
  local i = 0
  local j, s
  if love.filesystem.getInfo("data/cards.txt", "file") then
    tCARDS = {}
    for line in love.filesystem.lines("data/cards.txt") do
      if string.find(line, "ID=") then
        i = i + 1
        tCARDS[i] = {}
        tCARDS[i].ID = tonumber(string.sub(line, 4))
        tCARDS[i].Name = "Unknown"
        tCARDS[i].Class = "Unknown"
        tCARDS[i].Habitat = "Unknown"
        tCARDS[i].Tag = ""
        tCARDS[i].Type = "Unknown"
        tCARDS[i].Image = iUNKNOWN
        tCARDS[i].FaceUp = true
        tCARDS[i].Tier = 0
        tCARDS[i].MaxHealth = 0
        tCARDS[i].Wounds = 0
        tCARDS[i].Power = 0
        tCARDS[i].Special = "None"
        tCARDS[i].SpDesc = "None"
        tCARDS[i].SpVal = 0
        tCARDS[i].SpTargets = 0
        tCARDS[i].Maximum = 1
        tCARDS[i].Copies = 0
        tCARDS[i].Price = 0
        tCARDS[i].Front = iCARDBACK
        tCARDS[i].Back = iCARDBACK
        tCARDS[i].Unlocked = "Locked"
        tCARDS[i].Rating = 10
        tCARDS[i].Duplicate = 1
        tCARDS[i].Nocturnal = false
        tCARDS[i].Effect = {}
        for j = 1, #tEFFECTS, 1 do
          table.insert(tCARDS[i].Effect, 0)
        end
        tCARDS[i].FloatyText = {}
      elseif string.find(line, "Name=") then tCARDS[i].Name = string.sub(line,6)
      elseif string.find(line, "Class=") then tCARDS[i].Class = string.sub(line,7) 
      elseif string.find(line, "Habitat=") then
        tCARDS[i].Habitat = string.sub(line,9)
        if love.filesystem.getInfo("backgrounds/"..tCARDS[i].Habitat..".png", "file") then
          tCARDS[i].Front = love.graphics.newImage("backgrounds/"..tCARDS[i].Habitat..".png") end
      elseif string.find(line, "Tag=") then tCARDS[i].Tag = string.sub(line,5) 
      elseif string.find(line, "Type=") then tCARDS[i].Type = string.sub(line,6) 
      elseif string.find(line, "Pic=") then 
        j = math.floor(tonumber(string.sub(line,5)))
        if j > 0 then
          if j<10 then s = "images/pic00"..j..".jpg"
          elseif j<100 then s = "images/pic0"..j..".jpg"
          else s = "images/pic"..j..".jpg" end
          if love.filesystem.getInfo(s, "file") then tCARDS[i].Image = love.graphics.newImage(s) end
        end
      elseif string.find(line, "Tier=") then tCARDS[i].Tier = tonumber(string.sub(line,6)) 
      elseif string.find(line, "MaxHealth=") then tCARDS[i].MaxHealth = tonumber(string.sub(line,11)) 
      elseif string.find(line, "Power=") then tCARDS[i].Power = tonumber(string.sub(line,7))
      elseif string.find(line, "Special=") then tCARDS[i].Special = string.sub(line,9)
      elseif string.find(line, "SpDesc=") then tCARDS[i].SpDesc = string.sub(line,8)
      elseif string.find(line, "SpVal=") then tCARDS[i].SpVal = tonumber(string.sub(line,7))
      elseif string.find(line, "SpTargets=") then tCARDS[i].SpTargets = tonumber(string.sub(line, 11))
      elseif string.find(line, "Maximum=") then tCARDS[i].Maximum = tonumber(string.sub(line,9))
      elseif string.find(line, "Copies=") then
        tCARDS[i].Unlocked = "Unlocked"
        tCARDS[i].Copies = tonumber(string.sub(line,8))
      elseif string.find(line, "Price=") then tCARDS[i].Price = tonumber(string.sub(line, 7))
      elseif string.find(line, "Rating=") then tCARDS[i].Rating = tonumber(string.sub(line, 8))
      elseif string.find(line, "Duplicate=") then tCARDS[i].Duplicate = tonumber(string.sub(line, 11))
      else end
      if string.find(line, "Nocturnal.") then tCARDS[i].Nocturnal = true end
    end
  end
end

function loadcards()
  local i, j
  if love.filesystem.getInfo("saves/cards.txt", "file") then
    for line in love.filesystem.lines("saves/cards.txt") do
      if next(tCARDS) ~= nil then
        for i = 1, #tCARDS, 1 do
          if string.find(line, "GOLD_COINS=") then gCOINS = tonumber(string.sub(line, 12))
          elseif string.find(line, tCARDS[i].Name.."=") then
            j = tonumber(string.sub(line, string.len(tCARDS[i].Name)+2))
            if j < 0 then tCARDS[i].Unlocked = "Locked"
            else
              tCARDS[i].Unlocked = "Unlocked"
              tCARDS[i].Copies = j
            end
          else end
        end
      end
    end
  end
end

function savecards()
  local i, file
  if next(tCARDS) ~= nil then
    if not love.filesystem.getInfo("saves", "directory") then love.filesystem.createDirectory("saves") end
    if love.filesystem.getInfo("saves/cards.txt", "file") then love.filesystem.remove("saves/cards.txt") end
    file = love.filesystem.newFile("saves/cards.txt")
    file:open("w")
    file:write("GOLD_COINS="..gCOINS.."\r\n")
    for i = 1, #tCARDS, 1 do
      if tCARDS[i].Unlocked == "Locked" then file:write(tCARDS[i].Name.."=-1\r\n")
      else file:write(tCARDS[i].Name.."="..tCARDS[i].Copies.."\r\n") end
    end
    file:close()
  end
  
end

function readhabitats()  -- Read habitats
  local i = 0
  if love.filesystem.getInfo("data/habitats.txt", "file") then
    tHABITAT = {}
    for line in love.filesystem.lines("data/habitats.txt") do
      if string.find(line, "Name=") then
        i = i + 1
        tHABITAT[i] = {}
        tHABITAT[i].Name = string.sub(line, 6)
        if tHABITAT[i].Name == "Home" then tHABITAT[i].Unlock = 3 else tHABITAT[i].Unlock = 0 end
        tHABITAT[i].Easy = 0
        tHABITAT[i].Medium = 0
        tHABITAT[i].Hard = 0
        tHABITAT[i].Image = iCARDBACK
        tHABITAT[i].Square = iTILE
        if love.filesystem.getInfo("habitats/"..tHABITAT[i].Name..".png", "file") then
          tHABITAT[i].Image = love.graphics.newImage("habitats/"..tHABITAT[i].Name..".png")
        end
        if love.filesystem.getInfo("tiles/"..tHABITAT[i].Name..".png", "file") then
          tHABITAT[i].Square = love.graphics.newImage("tiles/"..tHABITAT[i].Name..".png")
        end
      elseif string.find(line, "Easy=") then
        tHABITAT[i].Easy = tonumber(string.sub(line, 6))
      elseif string.find(line, "Medium=") then
        tHABITAT[i].Medium = tonumber(string.sub(line, 8))
      elseif string.find(line, "Hard=") then
        tHABITAT[i].Hard = tonumber(string.sub(line, 6))
      else end
    end
  end
end

function loadlevels()    -- Load unlocked states
  local i, j
  if love.filesystem.getInfo("saves/levels.txt", "file") then
    for line in love.filesystem.lines("saves/levels.txt") do
      if next(tHABITAT) ~= nil then
        for i = 1, #tHABITAT, 1 do
          if string.find(line, tHABITAT[i].Name.."=") then
            j = tonumber(string.sub(line, string.len(line)))
            tHABITAT[i].Unlock = j
          end
        end
      end
    end
  end
end

function savelevels()    -- Save unlocked states
  local i, file
  if next(tHABITAT) ~= nil then
    if not love.filesystem.getInfo("saves", "directory") then love.filesystem.createDirectory("saves") end
    if love.filesystem.getInfo("saves/levels.txt", "file") then love.filesystem.remove("saves/levels.txt") end
    file = love.filesystem.newFile("saves/levels.txt")
    file:open("w")
    for i = 1, #tHABITAT, 1 do
      file:write(tHABITAT[i].Name.."="..tHABITAT[i].Unlock.."\r\n")
    end
    file:close()
  end
end

function loaddeck(filename, pl)  -- Load specific saved deck, takes filename without .txt
  local c, id
  if love.filesystem.getInfo("decks/"..filename..".txt", "file") then
    tDECK[pl] = {}
    for line in love.filesystem.lines("decks/"..filename..".txt") do
      id = tonumber(line)
      if id > 0 then 
        c = getcardbyid(id)
        if c > 0 then 
          if tCARDS[c].Type == "Hero" then tHERO[pl] = deepcopy(tCARDS[c])
          else table.insert(tDECK[pl], deepcopy(tCARDS[c])) end
        end
      end
    end
    if filename~="Autosave" then addpopup(filename.." loaded.") end
  end
end

function savedeck(filename, pl)  -- Save specific saved deck, takes filename without .txt
  local i, file
  if next(tDECK[pl]) ~= nil and next(tHERO[pl])~=nil then
    if not love.filesystem.getInfo("decks", "directory") then love.filesystem.createDirectory("decks") end
    if love.filesystem.getInfo("decks/"..filename.."txt", "file") then love.filesystem.remove("decks/"..filename..".txt") end
    file = love.filesystem.newFile("decks/"..filename..".txt")
    file:open("w")
    file:write(tHERO[pl].ID.."\r\n")
    for i = 1, #tDECK[pl], 1 do
      file:write(tDECK[pl][i].ID.."\r\n")
    end
    file:close()
    if filename~="Autosave" then addpopup(filename.." saved.") end
  end
end

function erasedeck(filename)     -- Erase specific saved deck, takes filename without .txt
  if love.filesystem.getInfo("decks/"..filename..".txt") then
    love.filesystem.remove("decks/"..filename..".txt")
    addpopup(filename.." erased.")
  end
end

function refreshsavefiles()
  local i
  tSAVES = love.filesystem.getDirectoryItems("decks")
  if next(tSAVES) ~= nil then
    for i = #tSAVES, 1, -1 do  -- Chops off the .txt from the filename
      if string.find(tSAVES[i], ".DS") then table.remove(tSAVES, i)
      elseif tSAVES[i] == "Autosave.txt" then table.remove(tSAVES, i)
      else tSAVES[i] = string.sub(tSAVES[i], 1, -5) end
    end
    table.sort(tSAVES)
  end
end

function getcardbyid(id)
  local c = 0
  local i
  if next(tCARDS) ~= nil then
    for i = 1, #tCARDS, 1 do
      if tCARDS[i].ID == id then c = i end
    end
  end
  return c
end

function getcardbyname(name)
  local c = 0
  local i
  if next(tCARDS) ~= nil then
    for i = 1, #tCARDS, 1 do
      if tCARDS[i].Name == name then c = i end
    end
  end
  return c
end

function getlistbyhabitat(habitat)
  local list = {}
  local i
  if next(tCARDS)~=nil then
    for i = 1, #tCARDS, 1 do
      if tCARDS[i].Habitat == habitat then table.insert(list, tCARDS[i]) end
    end
  end
  return list
end

function blankcard(card, heal)  -- Used in newgame but also a couple of actions
  local i
  card.FaceUp = true
  card.Used = false
  card.SpecialUsed = false
  card.Target = false
  card.Hidden = false
  card.Zoomed = false
  card.X = 0
  card.Y = 0
  card.Angle = 0
  card.Animate = false
  card.AniCount = 0
  card.OriginalX = 0
  card.OriginalY = 0
  card.TargetX = 0
  card.TargetY = 0
  for i = 1, #tEFFECTS, 1 do
    card.Effect[i] = 0
  end
  if heal == true then
    card.Wounds = 0
  end
end

function newgame()  -- Shuffle decks, setup and deal initial cards
  local i	, j
  print("NEW GAME: "..tCONTROL[1].." vs "..tCONTROL[2])
  gTURN = math.random(2)
  gDAYNIGHT = math.random(10)
  for j = 1, 2, 1 do
    blankcard(tHERO[j], true)
    tHERO[j].X = xDECK[j]
    tHERO[j].Y = yDECK[j]
    tHERO[j].Location = "Hero"
    tHERO[j].Animate = true
    tHERO[j].AniCount = gANIMATION_TIME
    tHERO[j].OriginalX = xDECK[j]
    tHERO[j].OriginalY = yDECK[j]
    tHERO[j].TargetX = xHERO[j]
    tHERO[j].TargetY = yHERO[j]
    for i = 1, #tDECK[j], 1 do
      blankcard(tDECK[j][i], true)
      tDECK[j][i].X = xDECK[j]
      tDECK[j][i].Y = yDECK[j]
      tDECK[j][i].Location = "Deck"
      tDECK[j][i].FaceUp = false
      tDECK[j][i].Animate = false
      tDECK[j][i].AniCount = 0
      tDECK[j][i].OriginalX = xDECK[j]
      tDECK[j][i].OriginalY = yDECK[j]
      tDECK[j][i].TargetX = xDECK[j]
      tDECK[j][i].TargetY = yDECK[j]
    end
  end
  -- playsound(sSOUND[4])
  for i = 1, gSHUFFLE, 1 do
    shuffledeck(1)
    shuffledeck(2)
  end
  for i = 1, 2, 1 do
    for j = 1, 2, 1 do
      dealcard(i)
    end
    tBUDGET[i] = 0
    tPOINTS[i] = 0
  end
  dealcard(3-gTURN)
  gGAMEMODE = "Battle"
  --playsound(sSOUND[7])
  arrangehand(3-gTURN)
  newturn(gTURN)
end

function shuffledeck(pl)
  local i = math.random(1, #tDECK[pl])
  local j = math.random(1, #tDECK[pl])
  tDECK[pl][i], tDECK[pl][j] = deepcopy(tDECK[pl][j]), deepcopy(tDECK[pl][i])
end

function addcardtodeck(card, pl)
  if card.Type == "Hero" then
    tHERO[pl] = deepcopy(card)
  else
    if getdeckcount(pl) < gDECKSIZE then
      table.insert(tDECK[pl], deepcopy(card))
    end
  end
end

function removecardfromdeck(card, pl)
  local i = 0
  if card.Type == "Hero" then
    tHERO[pl] = {}
  elseif next(tDECK[pl])~=nil then
    repeat
      i = i + 1
    until tDECK[pl][i].ID == card.ID or i == #tDECK[pl]
    if tDECK[pl][i].ID == card.ID then table.remove(tDECK[pl], i) end
  else end
end

function sortdeck(pl)
  local i, c, best, best_id
  local ordered, sorted = {}, {}
  if next(tDECK[pl])~=nil then
    for i = 1, #tDECK[pl], 1 do
      sorted[i] = false
    end
    repeat
      best = 999999
      best_id = 0
      for i = 1, #tDECK[pl], 1 do
        c = getcardbyid(tDECK[pl][i].ID)
        if c < best and sorted[i] == false then
          best = c
          best_id = i
        end
      end
      if best_id ~= 0 then
        table.insert(ordered, deepcopy(tDECK[pl][best_id]))
        sorted[best_id]=true
      end
    until best_id==0
    tDECK[pl] = deepcopy(ordered)
  end
end

function autodeck(pl, habitat)
  local list = {}
  local i, j, loops
  local best, best_ref
  local count, remaining, rating, duplicate, balance, cards
  if next(tCARDS)~=nil then 
    if next(tHERO[pl])==nil then -- can't select a hero if they already have one
      for i = 1, #tCARDS, 1 do   -- make a list of available heroes
        if tCARDS[i].Type == "Hero" and (tCARDS[i].Habitat == habitat or habitat == "All") then table.insert(list, i) end
      end
      if next(list)~=nil then
        if tCONTROL[pl]=="Player" or gDIFFICULTY==3 then loops=12
        elseif gDIFFICULTY==2 then loops=5
        else loops=2 end
        best, best_ref = -1, 0
        for i = 1, loops, 1 do   -- keep looping through the list looking for a better hero
          j = math.random(#list)
          if tCARDS[list[j]].Rating>best then
            best = tCARDS[list[j]].Rating
            best_ref = j
          end
        end
        if best_ref~=0 then addcardtodeck(tCARDS[list[best_ref]], pl) end  -- add the hero to the deck
      end
      list = {}
    end
    cards = getdeckcount(pl)
    if cards<gDECKSIZE then
      for i = 1, #tCARDS, 1 do   -- make a list of available cards (allies & actions)
        count = getcountindeckbyid(tCARDS[i].ID, pl)
        if tCONTROL[pl]=="AI" then remaining = tCARDS[i].Maximum - count 
        else remaining = tCARDS[i].Copies - count end
        if (tCARDS[i].Type=="Ally" or tCARDS[i].Type=="Action") and (tCARDS[i].Habitat==habitat or habitat=="All") and remaining>0 then
          for j = 1, remaining, 1 do
            table.insert(list, i)
          end
        end
      end
      if next(list)~=nil then
        if tCONTROL[pl]=="Player" or gDIFFICULTY==3 then loops=500
        elseif gDIFFICULTY==2 then loops=50
        else loops=5 end
        repeat
          best, best_ref = -1, 0
          for i = 1, loops, 1 do  -- keep looping through the cards looking for a better card
            j = math.random(#list)
            count = getcountindeckbyid(tCARDS[list[j]].ID, pl)
            duplicate = tCARDS[list[j]].Duplicate ^ count
            if cards > 0 then balance = 1 - (getcountindeckbytype(tCARDS[list[j]].Type, pl) / cards)
            else balance = 1 end
            rating = tCARDS[list[j]].Rating * duplicate * balance
            if rating>best then
              best = rating
              best_ref = j
            end
          end  
          if best_ref ~= 0 then 
            addcardtodeck(tCARDS[list[best_ref]], pl)  -- add the card to the deck
            table.remove(list, best_ref)               -- remove the card from the list
            cards = cards + 1                          -- use this counter rather than looping every time!
          end  
        until cards==gDECKSIZE
      end
    end
  end
end

function dealcard(pl)
  local size = #tHAND[pl]
  if size < gMAXCARDSINHAND then 
    if next(tDECK[pl]) ~= nil then
      table.insert(tHAND[pl], deepcopy(tDECK[pl][1]))
      tHAND[pl][size+1].Location = "Hand"
      tHAND[pl][size+1].Animate = true
      tHAND[pl][size+1].AniCount = gANIMATION_TIME
      tHAND[pl][size+1].OriginalX = xDECK[pl]
      tHAND[pl][size+1].OriginalY = yDECK[pl]
      tHAND[pl][size+1].TargetX = xHAND[pl]
      tHAND[pl][size+1].TargetY = yHAND[pl]
      if gTURN == pl and tCONTROL[pl] == "Player" then
        tHAND[pl][size+1].FaceUp = true
        tHAND[pl][size+1].Clickable = true
      end
      table.remove(tDECK[pl],1)
    else 
      -- The player loses 10 health
    end
  end
end

function newturn(pl)  -- deal cards, temporary effects end
  local a = pl      -- friendly
  local b = 3 - pl  -- foe
  local i, j
  local removed
  print("Start Player"..a.." turn")
  gDAYNIGHT = gDAYNIGHT + 1
  gCEASEFIRE = math.max(gCEASEFIRE-1, 0)
  if gDAYNIGHT>10 then gDAYNIGHT = 1 end
  if #tHAND[a] < gMAXCARDSINHAND then dealcard(a) end
  if #tHAND[a] < gMAXCARDSINHAND then dealcard(a) end
  tPOINTS[a] = tBUDGET[a]
  if next(tHERO[a]) ~= nil then
    tHERO[a].FaceUp = true
    tHERO[a].Clickable = true
    for j = 1, #tEFFECTS, 1 do
      if tHERO[a].Effect[j]<100 then tHERO[a].Effect[j] = math.max(tHERO[a].Effect[j]-1, 0) end
    end
  end
  if next(tHERO[b]) ~= nil then
    tHERO[b].FaceUp = true
    tHERO[b].Clickable = true
    tHERO[b].Used = false
    tHERO[b].SpecialUsed = false
    for j = 1, #tEFFECTS, 1 do
      if tHERO[b].Effect[j]<100 then tHERO[b].Effect[j] = math.max(tHERO[b].Effect[j]-1, 0) end
    end
  end
  if next(tDECK[a]) ~= nil then
    for i = #tDECK[a], 1, -1 do
      tDECK[a][i].FaceUp = false
      tDECK[a][i].Clickable = false
      for j = 1, #tEFFECTS, 1 do
        if tDECK[a][i].Effect[j]<100 then tDECK[a][i].Effect[j] = math.max(tDECK[a][i].Effect[j]-1, 0) end
      end
    end
  end
  if next(tDECK[b]) ~= nil then
    for i = #tDECK[b], 1, -1 do
      tDECK[b][i].FaceUp = false
      tDECK[b][i].Clickable = false
      for j = 1, #tEFFECTS, 1 do
        if tDECK[b][i].Effect[j]<100 then tDECK[b][i].Effect[j] = math.max(tDECK[b][i].Effect[j]-1, 0) end
      end
    end    
  end
  if next(tCOMBAT[a]) ~= nil then
    for i = #tCOMBAT[a], 1, -1 do
      removed = false
      tCOMBAT[a][i].FaceUp = true
      tCOMBAT[a][i].Clickable = true
      for j = 1, #tEFFECTS, 1 do
        if tCOMBAT[a][i].Effect[j]<100 then tCOMBAT[a][i].Effect[j] = math.max(tCOMBAT[a][i].Effect[j]-1, 0) end
      end
      if tCOMBAT[a][i].Nocturnal == true then
        if gDAYNIGHT>=6 then tCOMBAT[a][i].Effect[10] = 1
        elseif gDAYNIGHT~=0 then tCOMBAT[a][i].Effect[9] = 1
        else end
      end
      if tCOMBAT[a][i].Special == "Forgetful" and math.random(100)<=15 and gethandcount(a)<gMAXCARDSINHAND then
        addfloatytext(tCOMBAT[a][i], "-Forgetful-")
        if tCONTROL[a]=="AI" then tCOMBAT[a][i].FaceUp = false end
        table.insert(tHAND[a], deepcopy(tCOMBAT[a][i]))
        table.remove(tCOMBAT[a], i)
        removed = true
        arrangecombat(a)
        arrangehand(a)
      elseif tCOMBAT[a][i].Special == "Lazy" and math.random(100)<=15 then
        tCOMBAT[a][i].Effect[1] = math.max(tCOMBAT[a][i].Effect[1], 1)
        addfloatytext(tCOMBAT[a][i], "-Lazy-")
      elseif tCOMBAT[a][i].Special == "Breeds" and math.random(100)<=15 and #tCOMBAT[a] < gMAXCARDSINCOMBAT then
        addfloatytext(tCOMBAT[a][i], "+Breeds+")
        table.insert(tCOMBAT[a], deepcopy(tCOMBAT[a][i]))
        tCOMBAT[a][#tCOMBAT[a]].Wounds = 0
        arrangecombat(a)
      else end
      if tCOMBAT[a][i].Effect[11]>0 and removed == false then
        addfloatytext(tCOMBAT[a][i], "-Poison-")
        damage(tCOMBAT[a][i], "Combat", a, i, 1+getarmour(tCOMBAT[a][i]))
      elseif tCOMBAT[a][i].Effect[12]>0 and removed == false then
        addfloatytext(tCOMBAT[a][i], "-Poison-")
        damage(tCOMBAT[a][i], "Combat", a, i, 3+getarmour(tCOMBAT[a][i]))
      elseif tCOMBAT[a][i].Effect[13]>0 and removed == false then
        addfloatytext(tCOMBAT[a][i], "-Poison-")
        damage(tCOMBAT[a][i], "Combat", a, i, 5+getarmour(tCOMBAT[a][i]))
      else end
    end
  end
  if next(tCOMBAT[b]) ~= nil then
    for i = #tCOMBAT[b], 1, -1 do
      tCOMBAT[b][i].FaceUp = true
      tCOMBAT[b][i].Clickable = true
      tCOMBAT[b][i].Used = false
      tCOMBAT[b][i].SpecialUsed = false
      for j = 1, #tEFFECTS, 1 do
        if tCOMBAT[b][i].Effect[j]<100 then tCOMBAT[b][i].Effect[j] = math.max(tCOMBAT[b][i].Effect[j]-1, 0) end
      end
      if tCOMBAT[b][i].Nocturnal == true then
        if gDAYNIGHT>=6 then tCOMBAT[b][i].Effect[10] = 1
        elseif gDAYNIGHT~=0 then tCOMBAT[b][i].Effect[9] = 1
        else end
      end
    end
  end
  if next(tHAND[a]) ~= nil then
    for i = #tHAND[a], 1, -1 do
      if tCONTROL[a]=="Player" then tHAND[a][i].FaceUp = true end
      if tCONTROL[a]=="Player" then tHAND[a][i].Clickable = true end
      for j = 1, #tEFFECTS, 1 do
        if tHAND[a][i].Effect[j]<100 then tHAND[a][i].Effect[j] = math.max(tHAND[a][i].Effect[j]-1, 0) end
      end
    end
  end
  if next(tHAND[b]) ~= nil then
    for i = #tHAND[b], 1, -1 do
      tHAND[b][i].FaceUp = false
      tHAND[b][i].Clickable = false
      for j = 1, #tEFFECTS, 1 do
        if tHAND[b][i].Effect[j]<100 then tHAND[b][i].Effect[j] = math.max(tHAND[b][i].Effect[j]-1, 0) end
      end
    end
  end
  if next(tGRAVE[a]) ~= nil then
    for i = #tGRAVE[a], 1, -1 do
      tGRAVE[a][i].FaceUp = true
      tGRAVE[a][i].Clickable = false
      for j = 1, #tEFFECTS, 1 do
        if tGRAVE[a][i].Effect[j]<100 then tGRAVE[a][i].Effect[j] = math.max(tGRAVE[a][i].Effect[j]-1, 0) end
      end
    end
    tGRAVE[a][#tGRAVE[a]].Clickable = true
  end
  if next(tGRAVE[b]) ~= nil then
    for i = #tGRAVE[b], 1, -1 do
      tGRAVE[b][i].FaceUp = true
      tGRAVE[b][i].Clickable = false
      for j = 1, #tEFFECTS, 1 do
        if tGRAVE[b][i].Effect[j]<100 then tGRAVE[b][i].Effect[j] = math.max(tGRAVE[b][i].Effect[j]-1, 0) end
      end
    end
    tGRAVE[b][#tGRAVE[b]].Clickable = true
  end
  if tCONTROL[a] == "Player" and tCONTROL[b] == "Player" then gGAMEMODE = "New Turn"
  elseif tCONTROL[a] == "AI" then
    i = AIscore(a)
    AIlastmove = love.timer.getTime()
    print("INITIAL BOARDSCORE: "..i)
    AIbest_score = i
    AIbest_start = {}
    AIloop = 0
  else end
  tCLICK = {}
  gZOOMED = false
  gSACRIFICED = false
  gACTION = ""
  AIsimulation=false
  arrangehand(pl)
end

function animatecard(card)
  card.X = card.X + ((card.TargetX - card.OriginalX) / gANIMATION_TIME)
  card.Y = card.Y + ((card.TargetY - card.OriginalY) / gANIMATION_TIME)
  card.AniCount = card.AniCount - 1
  if card.AniCount <= 0 then
    --make card placement sound
    card.AniCount = 0
    card.X = card.TargetX
    card.Y = card.TargetY
    card.Animate = false
  end  
end

function animatetext(card)
  local i = 0
  repeat
    i = i + 1
    card.FloatyText[i].Timer = card.FloatyText[i].Timer + 1
  until i==#card.FloatyText or card.FloatyText[i].Timer <= 40
  if card.FloatyText[1].Timer >= gFLOAT_TIMER then table.remove(card.FloatyText, 1) end
end

function calculatemoves(pl)  -- what moves can the player make?
  local moves = {false, false, false, false, false, false} 
  local source = tCLICK[1].Source
  local card = {}
  if source == "Hand" then card = tHAND[tCLICK[1].Player][tCLICK[1].Index]
  elseif source == "Combat" then card = tCOMBAT[tCLICK[1].Player][tCLICK[1].Index]
  elseif source == "Grave" then card = tGRAVE[tCLICK[1].Player][tCLICK[1].Index]
  elseif source == "Deck" then card = tDECK[tCLICK[1].Player][tCLICK[1].Index]
  elseif source == "Hero" then card = tHERO[tCLICK[1].Player]
  else end
  -- 1=March (play), 2=Action (play), 3=Attack, 4=Use (special), 5=Target, 6=Sacrifice
  if next(card)~=nil then
    if source=="Hand" and tPOINTS[pl]>=card.Tier and card.Type=="Ally" and gTURN==pl and gACTION=="" and getcombatcount(pl)<gMAXCARDSINCOMBAT then moves[1] = true
    elseif source=="Hand" and tPOINTS[pl]>=card.Tier and card.Type=="Action" and gTURN==pl and gACTION=="" then moves[2] = true
    elseif (source=="Combat" or source=="Hero") and card.Used==false and gTURN==pl and gACTION=="" and getpower(card)>0 and card.Effect[1]==0 and card.Effect[3]==0 and gCEASEFIRE==0 then moves[3] = true
    else end
    if (source=="Combat" or source=="Hero") and tPOINTS[pl]>=card.SpVal and card.SpVal>0 and card.SpecialUsed==false and gTURN==pl and gACTION=="" then moves[4] = true end
    if card.Target==true and gACTION~="" then moves[5] = true end
    if source=="Hand" and gSACRIFICED==false and gTURN==pl and gACTION=="" then moves[6] = true end
  end
  return moves
end

function findtargets()  -- which cards can be targetted for this action?
  local a,b = gTURN, 3-gTURN
  local found = false
  if gACTION ~= "" then
    if next(tDECK[a]) ~= nil then
      for i = 1, #tDECK[a], 1 do tDECK[a][i].Target = false end
    end
    if next(tHAND[a]) ~= nil then
      for i = 1, #tHAND[a], 1 do tHAND[a][i].Target = false end
    end
    if next(tCOMBAT[a]) ~= nil then
      for i = 1, #tCOMBAT[a], 1 do
        tCOMBAT[a][i].Target = false
        if gACTION == "Cage" and tCOMBAT[a][i].Tier>=3 then
          tCOMBAT[a][i].Target = true
          found = true
        elseif gACTION == "Calm Mammal" and tCOMBAT[a][i].Class=="Mammal" and tCOMBAT[a][i].Effect[4]>0 then
          tCOMBAT[a][i].Target = true
          found = true
        else end
      end
    end
    if next(tGRAVE[a]) ~= nil then
      for i = 1, #tGRAVE[a], 1 do tGRAVE[a][i].Target = false end
    end
    if next(tHERO[a]) ~= nil then 
      tHERO[a].Target = false
    end
    if next(tDECK[b]) ~= nil then
      for i = 1, #tDECK[b], 1 do tDECK[b][i].Target = false end
    end
    if next(tHAND[b]) ~= nil then
      for i = 1, #tHAND[b], 1 do 
        tHAND[b][i].Target = false
        if gACTION == "Snatch Card" then tHAND[b][i].Target = true end
      end
    end
    if next(tCOMBAT[b]) ~= nil then
      for i = 1, #tCOMBAT[b], 1 do
        tCOMBAT[b][i].Target = false
        if gACTION == "Attack" and tCOMBAT[b][i].Effect[2]==0 then 
          tCOMBAT[b][i].Target = true
          found = true
        elseif gACTION == "Cage" and tCOMBAT[b][i].Tier>=3 and tCOMBAT[b][i].Effect[2]==0 then
          tCOMBAT[b][i].Target = true
          found = true
        elseif gACTION == "Charm Mammal" and tCOMBAT[b][i].Class=="Mammal" and tCOMBAT[b][i].Effect[2]==0 then
          tCOMBAT[b][i].Target = true
          found = true
        elseif gACTION == "Kitten Scratch" and tCOMBAT[b][i].Effect[2]==0 then
          tCOMBAT[b][i].Target = true
          found = true
        elseif gACTION == "Mesmerise" and tCOMBAT[b][i].Effect[2]==0 then
          tCOMBAT[b][i].Target = true
          found = true
        elseif gACTION == "Night Vision" and tCOMBAT[b][i].Effect[2]>0 then
          tCOMBAT[b][i].Target = true
          found = true
        elseif gACTION == "Paralysis (Minor)" and tCOMBAT[b][i].Tier<=2 and tCOMBAT[b][i].Effect[2]==0 then
          tCOMBAT[b][i].Target = true
          found = true
        elseif gACTION == "Poison (Minor)" and tCOMBAT[b][i].Effect[2]==0 then
          tCOMBAT[b][i].Target = true
          found = true
        elseif gACTION == "Snake Coil" and tCOMBAT[b][i].Effect[2]==0 and (tCOMBAT[b][i].Class == "Bird" or tCOMBAT[b][i].Class == "Mammal") then
          tCOMBAT[b][i].Target = true
          found = true
        else end
      end
    end
    if next(tGRAVE[b]) ~= nil then
      for i = 1, #tGRAVE[b], 1 do tGRAVE[b][i].Target = false end
    end
    if next(tHERO[b]) ~= nil then
      tHERO[b].Target = false
      if gACTION == "Attack" then 
        tHERO[b].Target = true
        found = true
      elseif gACTION == "Kitten Scratch" and tHERO[b].Effect[2]==0 then
        tHERO[b].Target = true
        found = true
      elseif gACTION == "Night Vision" and tHERO[b].Effect[2] > 0 then
        tHERO[b].Target = true
        found = true
      else end
    end
  end
  return found
end

function cleartargets()
  local i, j
  for i = 1, 2, 1 do
    if next(tHAND[i]) ~= nil then
      for j = 1, #tHAND[i], 1 do tHAND[i][j].Target = false end
    end
    if next(tDECK[i]) ~= nil then
      for j = 1, #tDECK[i], 1 do tDECK[i][j].Target = false end
    end
    if next(tCOMBAT[i]) ~= nil then
      for j = 1, #tCOMBAT[i], 1 do tCOMBAT[i][j].Target = false end
    end
    if next(tGRAVE[i]) ~= nil then
      for j = 1, #tGRAVE[i], 1 do tGRAVE[i][j].Target = false end
    end
    if next(tHERO[i]) ~= nil then tHERO[i].Target = false end
  end
end

function dotarget()  -- perform action on target (e.g. attack, play action)
  local a, b = tCLICK[2].Player, tCLICK[1].Player
  local origin, target
  local i
  if tCLICK[2].Source == "Hand" then origin = tHAND[a][tCLICK[2].Index]
  elseif tCLICK[2].Source == "Combat" then origin = tCOMBAT[a][tCLICK[2].Index]
  elseif tCLICK[2].Source == "Grave" then origin = tGRAVE[a][tCLICK[2].Index]
  elseif tCLICK[2].Source == "Deck" then origin = tDECK[a][tCLICK[2].Index]
  elseif tCLICK[2].Source == "Hero" then origin = tHERO[a]
  else end
  if tCLICK[1].Source == "Hand" then target = tHAND[b][tCLICK[1].Index]
  elseif tCLICK[1].Source == "Combat" then target = tCOMBAT[b][tCLICK[1].Index]
  elseif tCLICK[1].Source == "Grave" then target = tGRAVE[b][tCLICK[1].Index]
  elseif tCLICK[1].Source == "Deck" then target = tDECK[b][tCLICK[1].Index]
  elseif tCLICK[1].Source == "Hero" then target = tHERO[b]
  else end
  if gACTION == "Attack" then
    dobattle(origin, tCLICK[2].Source, a, tCLICK[2].Index, target, tCLICK[1].Source, b, tCLICK[1].Index)
    if origin.Special == "Energetic" and math.random(100)<=20 then
      addfloatytext(origin, "+Energetic+")
    else
      origin.Used = true
    end
  elseif gACTION == "Cage" then
    target.Effect[1] = math.max(target.Effect[1], 4)
    target.Effect[2] = math.max(target.Effect[2], 4)
    addpopup(target.Name.." is caged")
    addfloatytext(target, "-Caged-")
  elseif gACTION == "Calm Mammal" then
    target.Effect[1] = 0
    target.Effect[4] = 0
    addpopup(target.Name.." is calmed")
    addfloatytext(target, "+Calm+")
  elseif gACTION == "Charm Mammal" then
    target.Effect[1] = math.max(target.Effect[1], 2)
    addpopup(target.Name.." is charmed")
    addfloatytext(target, "-Charmed-")
  elseif gACTION == "Kitten Scratch" then
    addpopup(target.Name.." is scratched")
    addfloatytext(target, "-Scratch-")
    damage(target, tCLICK[1].Source, b, tCLICK[1].Index, 1)
  elseif gACTION == "Mesmerise" then
    target.Effect[1] = math.max(target.Effect[1], 2)
    addpopup(target.Name.." is mesmerised")
    addfloatytext(target, "-Distracted-")    
  elseif gACTION == "Paralysis (Minor)" then
    target.Effect[3] = 4
    addpopup(target.Name.." is paralysed")
    addfloatytext(target, "-Paralysed-")
  elseif gACTION == "Poison (Minor)" then
    target.Effect[11] = 100
    addpopup(target.Name.." is poisoned")
    addfloatytext(target, "-Poisoned-")
  elseif gACTION == "Snake Coil" then
    if math.random(100)<=75 then
      target.Effect[3] = 4
      addpopup(target.Name.." is trapped")
      addfloatytext(target, "-Trapped-")
    else
      addpopup(target.Name.." escapes")
      addfloatytext(target, "+Escaped+")
    end
  elseif gACTION == "Snatch Card" then
    if math.random(100)<=75 then
      addpopup(origin.Name.." snatches "..target.Name)
      target.Animate = true
      target.AniCount = gANIMATION_TIME
      target.OriginalX = target.X
      target.OriginalY = target.Y
      target.Hover = false
      target.Angle = 0
      if gethandcount(a) < gMAXCARDSINHAND then        
        target.TargetX = xHAND[a]
        target.TargetY = yHAND[a]
        if gTURN == a and tCONTROL[a] == "Player" then
          target.FaceUp = true
          target.Clickable = true
        else 
          target.FaceUp = false
          target.Clickable = false
        end
        table.insert(tHAND[a], deepcopy(target))
        arrangehand(a)
        table.remove(tHAND[b], tCLICK[1].Index)
        arrangehand(b)
      else
        target.TargetX = xGRAVE[a]
        target.TargetY = yGRAVE[a]
        target.FaceUp = true
        target.Clickable = true
        for i = 1, #tGRAVE[a], 1 do
          tGRAVE[a][i].Clickable = false 
        end
        table.insert(tGRAVE[a], deepcopy(target))
        table.remove(tHAND[b], tCLICK[1].Index)
        arrangehand(b)
      end
    else addpopup(origin.Name.." fails to snatch a card") end
  else end
end

function dobattle(att_card, att_source, att_player, att_index, def_card, def_source, def_player, def_index)  -- battle between two cards
  local att_power = getpower(att_card)
  local def_power = getpower(def_card)
  addpopup(att_card.Name.." attacks "..def_card.Name)
  att_card.Effect[2] = 0
  if att_power>0 then
    if def_card.Special == "Evasive" and math.random(100)<=25 then
      addfloatytext(def_card, "+Evaded+")
    elseif def_card.Special == "Imitate Call" and math.random(100)<=50 and (att_card.Class=="Bird" or att_card.Class=="Mammal") then
      addfloatytext(def_card, "+Imitate+")
    elseif def_card.Special == "Trickery" and math.random(100)<=50 then
      addfloatytext(def_card, "+Trickery+")
    else
      damage(def_card, def_source, def_player, def_index, att_power)
    end
  end
  if def_power>0 and def_card.Effect[3]==0 then
    if att_card.Special == "Evasive" and math.random(100)<=25 then
      addfloatytext(att_card, "+Evaded+")
    else
      damage(att_card, att_source, att_player, att_index, def_power)
    end
  end
  if att_card ~= nil then
    if att_card.Special == "Single Use" then
      addfloatytext(att_card, "-Single Use-")
      damage(att_card, att_source, att_player, att_index, getkilldmg(att_card))
    end
  end
  if def_card ~= nil then
    if def_card.Special == "Single Use" then
      addfloatytext(def_card, "-Single Use-")
      damage(def_card, def_source, def_player, def_index, getkilldmg(def_card))
    end
  end
end

function damage(card, loc, pl, index, amount)
  local i
  if string.find(card.Special, "Armour") then amount = amount - tonumber(string.sub(card.Special, 9)) end
  amount = math.max(amount, 0)
  if amount > 0 then addfloatytext(card, "-"..amount)
  else addfloatytext(card, "0") end
  card.Wounds = card.Wounds + amount
  if AIsimulation==false then print(card.Name.." suffers "..amount.." damage") end
  if gethealth(card)<=0 then
    if AIsimulation==false then print(card.Name.." dies") end
    if math.random(100)<=50 and card.Special=="Regrow" then
      card.Wounds = 0
      addfloatytext(card, "+Regrows+")
      card.Effect[3] = math.max(card.Effect[3], 2)
    else 
      sendgraveyard(card, loc, pl, index)
      if next(tCOMBAT[pl])~=nil then
        for i = 1, #tCOMBAT[pl], 1 do
          if tCOMBAT[pl][i].Special=="Carrion" and tCOMBAT[pl][i].Wounds>0 then
            addfloatytext(tCOMBAT[pl][i], "+Carrion+")
            tCOMBAT[pl][i].Wounds = tCOMBAT[pl][i].Wounds - 1
          end
        end
      end
    end
    return true
  else
    return false
  end
end

function sendgraveyard(card, loc, pl, index)  -- send card to graveyard pile
  local i
  card.Used = false
  card.SpecialUsed = false
  card.Animate = true
  card.AniCount = gANIMATION_TIME
  card.OriginalX = card.X
  card.OriginalY = card.Y
  card.TargetX = xGRAVE[pl]
  card.TargetY = yGRAVE[pl]
  card.FaceUp = true
  card.Hover = false
  card.Angle = 11 - math.random(20)
  for i = 1, #tEFFECTS, 1 do card.Effect[i] = 0 end
  table.insert(tGRAVE[pl], deepcopy(card))
  for i = 1, #tGRAVE[pl], 1 do
    tGRAVE[pl][i].Clickable = false
  end
  tGRAVE[pl][#tGRAVE[pl]].Clickable = true
  if loc == "Combat" then
    --if card.sDeath ~= 0 then makesound(sSOUND[card.sDeath]) end
    addfloatytext(card, "-death-")
    table.remove(tCOMBAT[pl], index)
    arrangecombat(pl)
  elseif loc == "Hand" then
    table.remove(tHAND[pl], index)
    arrangehand(pl)
  elseif loc == "Deck" then 
    table.remove(tDECK[pl], index)
  elseif loc == "Hero" and AIsimulation==false then
    gENDING = gFPS * 4
    addpopup("P"..3-pl.." has won the game")
  else end  -- Impossible to send from graveyard!
  --hidetargets(pl)
  --naturalbonus(pl)	
end

function senddeck(card, loc, pl, index)     -- send card to deck
  local i
  if AIsimulation==false then print(card.Name.." is sent to the deck") end
  card.Used = false
  card.SpecialUsed = false
  card.Animate = true
  card.AniCount = gANIMATION_TIME
  card.OriginalX = card.X
  card.OriginalY = card.Y
  card.TargetX = xDECK[pl]
  card.TargetY = yDECK[pl]
  card.FaceUp = false
  card.Clickable = false
  card.Hover = false
  card.Angle = 0
  for i = 1, #tEFFECTS, 1 do 
    if card.Effect[i] < 100 then card.Effect[i] = 0 end
  end
  table.insert(tDECK[pl], math.random(getdeckcount(pl)), deepcopy(card))
  if loc == "Combat" then
    table.remove(tCOMBAT[pl], index)
    arrangecombat(pl)
  elseif loc == "Hand" then
    table.remove(tHAND[pl], index)
    arrangehand(pl)
  elseif loc == "Grave" then
    table.remove(tGRAVE[pl], index)
  else end  -- Nothing happens from Hero and impossible to send from deck!
  --hidetargets(pl)
  --naturalbonus(pl)
end

function sendhand(card, loc, pl, index)    -- send card to hand
  local i
  if AIsimulation==false then print(card.Name.." is sent to player "..pl.."'s hand") end
  card.Used = false
  card.SpecialUsed = false
  card.Animate = true
  card.AniCount = gANIMATION_TIME
  card.OriginalX = card.X
  card.OriginalY = card.Y
  card.TargetX = xHAND[pl]
  card.TargetY = yHAND[pl]
  if gTURN == pl and tCONTROL[pl] == "Player" then 
    card.FaceUp = true
    card.Clickable = true
  else card.FaceUp = false end
  card.Hover = false
  card.Angle = 0
  for i = 1, #tEFFECTS, 1 do 
    if card.Effect[i] < 100 then card.Effect[i] = 0 end
  end
  table.insert(tHAND[pl], deepcopy(card))
  arrangehand(pl)
  if loc == "Combat" then
    table.remove(tCOMBAT[pl], index)
    arrangecombat(pl)
  elseif loc == "Deck" then
    table.remove(tDECK[pl], index)
  elseif loc == "Grave" then
    table.remove(tGRAVE[pl], index)
  else end  --  Nothing happens from Hero and impossible to send from hand!
  --hidetargets(pl)
  --naturalbonus(pl)
end

function sendcombat(card, loc, pl, index)   -- send card into combat
  local i
  local queen_ant = false
  --if card.sSummon ~= 0 then makesound(sSOUND[card.sSummon]) end
  card.Animate = true
  card.AniCount = gANIMATION_TIME
  card.OriginalX = card.X
  card.OriginalY = card.Y
  card.TargetX = xCOMBAT[player]
  card.TargetY = yCOMBAT[player]
  if card.Name == "Queen Ant" and next(tCOMBAT[pl])~=nil then           -- Play queen ant, soldiers stronger
    for i = 1, #tCOMBAT[pl], 1 do
      if tCOMBAT[pl][i].Name == "Soldier Ant" then
        addfloatytext(tCOMBAT[pl][i], "+Defend Queen+")
        tCOMBAT[pl][i].Power = tCOMBAT[pl][i].Power + 2
        tCOMBAT[pl][i].MaxHealth = tCOMBAT[pl][i].MaxHealth + 2
        tCOMBAT[pl][i].Effect[8] = 100
      end
    end
  elseif card.Name == "Soldier Ant" and next(tCOMBAT[pl])~=nil then     -- Play soldier  ant, queen already in play
    for i = 1, #tCOMBAT[pl], 1 do
      if tCOMBAT[pl][i].Name == "Queen Ant" then queen_ant = true end
    end
    if queen_ant == true then
      addfloatytext(card, "+Defend Queen+")
      card.Power = card.Power + 2
      card.MaxHealth = card.MaxHealth + 2
      card.Effect[8] = 100
    end
  else end
  if card.Special == "Industrious" then tPOINTS[pl] = math.min(tPOINTS[pl]+1, tBUDGET[pl])    -- Industrious, gain a spending point
  elseif card.Special == "Stealthy" then card.Effect[2] = 100             -- Stealthy, hidden target
  elseif card.Special == "Excavate" then                                  -- Excavate, draw 3 cards
    addpopup(card.Name.." digs a tunnel")
    if loc == "Hand" then table.remove(tHAND[pl], index) end            -- Ugly fix to give space for cards
    for i = 1, 3, 1 do
      dealcard(pl)
    end
    if loc == "Hand" then table.insert(tHAND[pl], index, deepcopy(card)) end   -- Ugly fix to delete it again later!
  else end
  if card.Special == "Scary to Mammals" and next(tCOMBAT[3-pl])~=nil then  -- Mammals scared of it
    for i = 1, #tCOMBAT[3-pl], 1 do
      if tCOMBAT[3-pl][i].Tier<=4 and tCOMBAT[3-pl][i].Class=="Mammal" then 
        if math.random(100)<=50 or tCOMBAT[3-pl][i].Special=="Fearful" then
          tCOMBAT[3-pl][i].Effect[4] = math.max(tCOMBAT[3-pl][i].Effect[4], 2)
          tCOMBAT[3-pl][i].Effect[1] = math.max(tCOMBAT[3-pl][i].Effect[1], 2)
          addfloatytext(tCOMBAT[3-pl][i], "-Fear-")
          addfloatytext(card, "+Scary+")
        else addfloatytext(tCOMBAT[3-pl][i], "+Resist+") end
      end
    end
  end
  if card.Class=="Mammal" and card.Tier<=4 and next(tCOMBAT[3-pl])~=nil then  -- Check for scary to mammals
    for i = 1, #tCOMBAT[3-pl], 1 do
      if tCOMBAT[3-pl][i].Special=="Scary to Mammals" and (math.random(100)<=50 or card.Special=="Fearful") then
        card.Effect[4] = math.max(card.Effect[4], 2)
        card.Effect[1] = math.max(card.Effect[1], 2)
        addfloatytext(card, "-Fear-")
        addfloatytext(tCOMBAT[3-pl][i], "+Scary+")
      end
    end
  end
  if card.Nocturnal == true then  -- Nocturnal bonus/penalty
    if gDAYNIGHT>=6 then card.Effect[10] = 1
    elseif gDAYNIGHT~=0 then card.Effect[9] = 1
    else end
  end
  if card.Special ~= "Surprise Attack" then card.Used = true end
  card.SpecialUsed = true
  card.FaceUp = true
  card.Clickable = true
  table.insert(tCOMBAT[pl], deepcopy(card))
  if loc == "Hand" then 
    table.remove(tHAND[pl], index)
    arrangehand(pl)
  elseif loc == "Grave" then table.remove(tGRAVE[pl], index)
  elseif loc == "Deck" then table.remove(tDECK[pl], index)
  else end
  arrangecombat(pl)
  -- hidetargets(player)
  -- naturalbonus(player)
end

function playaction(card, pl)  -- Play action or use special ability
  local sp = card.Special      -- Special ability
  local r = false              -- Flag that it was successful
  local a = pl                 -- Friendly
  local b = 3 - pl             -- Foe
  local targets = {}           -- List of potential targets to randomly pick from
  local i, j
  if card.SpTargets > 0 then   -- These actions have a target, same step for each
    if findtargets() == true then
      if card.SpTargets > 1 then addpopup("Choose targets")
      else addpopup("Choose target") end
      r = true
    else
      addpopup("No target available")
      gACTION = ""
    end
  else                         -- Now alphabetical, these actions have no target, do them!
    if sp == "Feed Birds" then
      addpopup("P"..a.." feeds the birds")
      if next(tCOMBAT[a])~=nil then
        for i = 1, #tCOMBAT[a], 1 do
          if tCOMBAT[a][i].Class == "Bird" then
            addfloatytext(tCOMBAT[a][i], "+Fed+")
            addfloatytext(tCOMBAT[a][i], "+"..math.min(tCOMBAT[a][i].Wounds,10).." Health")
            tCOMBAT[a][i].Wounds = math.max(tCOMBAT[a][i].Wounds-10, 0)
            tCOMBAT[a][i].Effect[5] = 100
            r = true
          end
        end
      end
    elseif sp == "Feed Fish" then
      addpopup("P"..a.." feeds the fish")
      if next(tCOMBAT[a])~=nil then
        for i = 1, #tCOMBAT[a], 1 do
          if tCOMBAT[a][i].Class == "Fish" then
            addfloatytext(tCOMBAT[a][i], "+Fed+")
            addfloatytext(tCOMBAT[a][i], "+"..math.min(tCOMBAT[a][i].Wounds,10).." Health")
            tCOMBAT[a][i].Wounds = math.max(tCOMBAT[a][i].Wounds-10, 0)
            tCOMBAT[a][i].Effect[5] = 100
            r = true
          end
        end
      end
    elseif sp == "Feed Insects" then
      addpopup("P"..a.." feeds the insects")
      if next(tCOMBAT[a])~=nil then
        for i = 1, #tCOMBAT[a], 1 do
          if tCOMBAT[a][i].Class == "Insect" then
            addfloatytext(tCOMBAT[a][i], "+Fed+")
            addfloatytext(tCOMBAT[a][i], "+"..math.min(tCOMBAT[a][i].Wounds,10).." Health")
            tCOMBAT[a][i].Wounds = math.max(tCOMBAT[a][i].Wounds-10, 0)
            tCOMBAT[a][i].Effect[5] = 100
            r = true
          end
        end
      end
    elseif sp == "Feed Mammals" then
      addpopup("P"..a.." feeds the mammals")
      if next(tCOMBAT[a])~=nil then
        for i = 1, #tCOMBAT[a], 1 do
          if tCOMBAT[a][i].Class == "Mammal" then
            addfloatytext(tCOMBAT[a][i], "+Fed+")
            addfloatytext(tCOMBAT[a][i], "+"..math.min(tCOMBAT[a][i].Wounds,10).." Health")
            tCOMBAT[a][i].Wounds = math.max(tCOMBAT[a][i].Wounds-10, 0)
            tCOMBAT[a][i].Effect[5] = 100
            r = true
          end
        end
      end
    elseif sp == "High Winds" then
      addpopup("P"..a.." summons a gusty gale")
      if next(tCOMBAT[a])~=nil then
        for i = #tCOMBAT[a], 1, -1 do
          if tCOMBAT[a][i].Tier<=2 then
            addfloatytext(tCOMBAT[a][i], "-Blown Away-")
            damage(tCOMBAT[a][i], "Combat", a, i, getkilldmg(tCOMBAT[a][i]))
          end
        end
      end
      if next(tCOMBAT[b])~=nil then
        for i = #tCOMBAT[b], 1, -1 do
          if tCOMBAT[b][i].Tier<=2 then
            addfloatytext(tCOMBAT[b][i], "-Blown Away-")
            damage(tCOMBAT[b][i], "Combat", b, i, getkilldmg(tCOMBAT[b][i]))
          end
        end
      end
      r = true
    elseif sp == "Hoover" then
      addpopup("Time for a spring clean!")
      r = true
      if next(tCOMBAT[a])~=nil then
        for i = #tCOMBAT[a], 1, -1 do
          if (tCOMBAT[a][i].Class == "Insect" or tCOMBAT[a][i].Class == "Arachnid") and tCOMBAT[a][i].Tier <= 5 then
            addfloatytext(tCOMBAT[a][i], "-Hoovered-")
            damage(tCOMBAT[a][i], "Combat", a, i, getkilldmg(tCOMBAT[a][i]))
          elseif tCOMBAT[a][i].Class ~= "Fish" and tCOMBAT[a][i].Tier <= 5 then
            addfloatytext(tCOMBAT[a][i], "-Fear-")
            tCOMBAT[a][i].Effect[4] = math.max(tCOMBAT[a][i].Effect[4], 2)
            tCOMBAT[a][i].Effect[1] = math.max(tCOMBAT[a][i].Effect[1], 2)
          else 
            addfloatytext(tCOMBAT[a][i], "+Unaffected+")
          end
        end
      end
      if next(tCOMBAT[b])~=nil then
        for i = #tCOMBAT[b], 1, -1 do
          if (tCOMBAT[b][i].Class == "Insect" or tCOMBAT[b][i].Class == "Arachnid") and tCOMBAT[b][i].Tier <= 5 then
            addfloatytext(tCOMBAT[b][i], "-Hoovered-")
            damage(tCOMBAT[b][i], "Combat", b, i, getkilldmg(tCOMBAT[b][i]))
          elseif tCOMBAT[b][i].Class ~= "Fish" and tCOMBAT[b][i].Tier <= 5 then
            addfloatytext(tCOMBAT[b][i], "-Fear-")
            tCOMBAT[b][i].Effect[4] = math.max(tCOMBAT[b][i].Effect[4], 2)
            tCOMBAT[b][i].Effect[1] = math.max(tCOMBAT[b][i].Effect[1], 2)
          else 
            addfloatytext(tCOMBAT[b][i], "+Unaffected+")
          end
        end
      end
    elseif sp == "Insecticide" then
      addpopup("It's armageddon for insects!")
      if next(tCOMBAT[a])~=nil then
        for i = #tCOMBAT[a], 1, -1 do
          addfloatytext(tCOMBAT[a][i], "-Sprayed-")
          damage(tCOMBAT[a][i], "Combat", a, i, getkilldmg(tCOMBAT[a][i]))
        end
      end
      if next(tCOMBAT[b])~=nil then
        for i = #tCOMBAT[b], 1, -1 do
          addfloatytext(tCOMBAT[b][i], "-Sprayed-")
          damage(tCOMBAT[b][i], "Combat", b, i, getkilldmg(tCOMBAT[b][i]))
        end
      end
      r = true
    elseif sp == "Intimidate" then
      addpopup(card.Name.." intimidates")
      if next(tCOMBAT[b])~=nil then
        for i = 1, #tCOMBAT[b], 1 do
          if tCOMBAT[b][i].Class == "Mammal" and tCOMBAT[b][i].Tier < card.Tier then
            addfloatytext(tCOMBAT[b][i], "-Fear-")
            tCOMBAT[b][i].Effect[1] = math.max(tCOMBAT[b][i].Effect[1], 2)
            tCOMBAT[b][i].Effect[4] = math.max(tCOMBAT[b][i].Effect[4], 2)
            r = true
          end
        end
      end
    elseif sp == "Mow Lawn" then
      if next(tCOMBAT[b])~=nil then
        for i = #tCOMBAT[b], 1, -1 do
          if tCOMBAT[b][i].Class=="Annelid" or tCOMBAT[b][i].Class=="Mammal" or tCOMBAT[b][i].Class=="Mollusc" then
            damage(tCOMBAT[b][i], "Combat", b, i, 5)
          end
        end
        r = true
      end
    elseif sp == "Night Vision" then
      if next(tCOMBAT[b])~=nil then
        for i = 1, #tCOMBAT[b], 1 do
          if tCOMBAT[b][i].Effect[2]>0 then table.insert(targets, i) end
        end
      end
      if next(tHERO[b])~=nil then
        if tHERO[b].Effect[2]>0 then table.insert(targets, 0) end
      end
      if next(targets)~=nil then
        addpopup("Target revealed")
        r = true
        j = math.random(#targets)
        i = targets[j]
        if i>0 then
          tCOMBAT[b][i].Effect[2]=0
          addfloatytext(tCOMBAT[b][i], "-Revealed-")
        else
          tHERO[b].Effect[2]=0
          addfloatytext(tHERO[b], "-Revealed-")
        end
      end
    elseif sp == "Rage" then
      addpopup(card.Name.." is enraged")
      card.Effect[14] = math.max(card.Effect[14], 1)
      r = true
    elseif sp == "Rain Delay" then
      addpopup("A torrential downpour")
      gCEASEFIRE = math.max(gCEASEFIRE, 3)
      r = true
    elseif sp == "Reinforcements" then
      addpopup("P"..a.." whistles for reinforcements")
      r = true
      dealcard(a)
      if gethandcount(a)<=gMAXCARDSINHAND then dealcard(a) end
      arrangehand(a)
    elseif sp == "Repel Insects" then
      addpopup("P"..a.." uses insect repellent")
      r = true
      if next(tCOMBAT[a])~=nil then
        for i = #tCOMBAT[a], 1, -1 do
          if tCOMBAT[a][i].Class=="Insect" and tCOMBAT[a][i].Tier<=3 then
            if math.random(100)<=75 then
              addfloatytext(tCOMBAT[a][i], "-Repelled-")
              senddeck(tCOMBAT[a][i], "Combat", a, i)
            else addfloatytext(tCOMBAT[a][i], "+Survived+") end
          end
        end
      end
      if next(tCOMBAT[b])~=nil then
        for i = #tCOMBAT[b], 1, -1 do
          if tCOMBAT[b][i].Class=="Insect" and tCOMBAT[b][i].Tier<=3 then
            if math.random(100)<=75 then
              addfloatytext(tCOMBAT[b][i], "-Repelled-")
              senddeck(tCOMBAT[b][i], "Combat", b, i)
            else addfloatytext(tCOMBAT[b][i], "+Survived+") end
          end
        end
      end
    elseif sp == "Resourceful" and gethandcount(a) < gMAXCARDSINHAND then
      addpopup("P"..a.." finds something useful")
      dealcard(a)
      arrangehand(a)
      r = true
    elseif sp == "Reveal Cards" then
      addpopup("P"..a.." takes a look at P"..b.."'s cards")
      if next(tHAND[b])~=nil then
        for i = 1, #tHAND[b], 1 do
          tHAND[b][i].FaceUp = true
          tHAND[b][i].Clickable = true
        end
        r = true
      end
    elseif sp == "Shine Light" then
      if gDAYNIGHT<=5 then
        if next(tCOMBAT[a])~=nil then
          for i = 1, #tCOMBAT[a], 1 do
            if tCOMBAT[a][i].Class == "Insect" then
              tCOMBAT[a][i].Effect[6] = math.max(tCOMBAT[a][i].Effect[6], 4)
              addfloatytext(tCOMBAT[a][i], "+Light+")
            elseif tCOMBAT[a][i].Nocturnal == true then 
              tCOMBAT[a][i].Effect[7] = math.max(tCOMBAT[a][i].Effect[7], 4)
              addfloatytext(tCOMBAT[a][i], "-Blinded-")
            else end
          end
        end
        if next(tCOMBAT[b])~=nil then
          for i = 1, #tCOMBAT[b], 1 do
            if tCOMBAT[b][i].Class == "Insect" then
              tCOMBAT[b][i].Effect[6] = math.max(tCOMBAT[b][i].Effect[6], 4)
              addfloatytext(tCOMBAT[b][i], "+Light+")
            elseif tCOMBAT[b][i].Nocturnal == true then 
              tCOMBAT[b][i].Effect[7] = math.max(tCOMBAT[b][i].Effect[7], 4)
              addfloatytext(tCOMBAT[b][i], "-Blinded-")
            else end
          end
        end
        r = true
        addpopup("A bright light")
      else
        addpopup("Use only  at night")
      end
    elseif sp == "Spray Water" then
      addpopup("P"..a.." sprays the garden hose!")
      if next(tCOMBAT[a])~=nil then
        for i = #tCOMBAT[a], 1, -1 do
          if tCOMBAT[a][i].Class=="Insect" then
            addfloatytext(tCOMBAT[a][i], "-Sprayed-")
            damage(tCOMBAT[a][i], "Combat", a, i, 2)
          elseif tCOMBAT[a][i].Class=="Fish" or tCOMBAT[a][i].Class=="Annelid" or tCOMBAT[a][i].Class=="Mollusc" or tCOMBAT[a][i].Class=="Reptile" or tCOMBAT[a][i].Class=="Crustacean" then
            addfloatytext(tCOMBAT[a][i], "+Sprayed+")
            tCOMBAT[a][i].Power = tCOMBAT[a][i].Power + 1
          else
            addfloatytext(tCOMBAT[a][i], "-Sprayed-")
            tCOMBAT[a][i].Power = math.max(tCOMBAT[a][i].Power-1, 0)
          end
        end
      end
      if next(tCOMBAT[b])~=nil then
        for i = #tCOMBAT[b], 1, -1 do
          if tCOMBAT[b][i].Class=="Insect" then
            addfloatytext(tCOMBAT[b][i], "-Sprayed-")
            damage(tCOMBAT[b][i], "Combat", a, i, 2)
          elseif tCOMBAT[b][i].Class=="Fish" or tCOMBAT[b][i].Class=="Annelid" or tCOMBAT[b][i].Class=="Mollusc" or tCOMBAT[b][i].Class=="Reptile" or tCOMBAT[b][i].Class=="Crustacean" then
            addfloatytext(tCOMBAT[b][i], "+Sprayed+")
            tCOMBAT[b][i].Power = tCOMBAT[b][i].Power + 1
          else
            addfloatytext(tCOMBAT[b][i], "-Sprayed-")
            tCOMBAT[b][i].Power = math.max(tCOMBAT[b][i].Power-1, 0)
          end
        end
      end
      r = true
    elseif sp == "Swarm" then
      addpopup("A swarm of bees descends upon P"..b)
      if next(tCOMBAT[b])~=nil then
        for i = #tCOMBAT[b], 1, -1 do
          damage(tCOMBAT[b][i], "Combat", b, i, 5)
        end
      end
      if next(tHERO[b])~=nil then
        damage(tHERO[b], "Hero", b, b, 5)
      end
      if next(tCOMBAT[a])~=nil then
        for i = #tCOMBAT[a], 1, -1 do
          if tCOMBAT[a][i].Name == "Drone Bee" or tCOMBAT[a][i].Name == "Queen Bee" then
            damage(tCOMBAT[a][i], "Combat", a, i, getkilldmg(tCOMBAT[a][i]))
          end
        end
      end
      r = true
    elseif sp == "Trip to the Vet" then
      addpopup("Oh no! It's the Veterinarian!")      
      if next(tCOMBAT[a])~=nil then
        r = true
        for i = #tCOMBAT[a], 1, -1 do
          if tCOMBAT[a][i].Class~="Insect" and tCOMBAT[a][i].Class~="Arachnid" and tCOMBAT[a][i].Habitat=="Home" then
            tCOMBAT[a][i].Wounds = 0
            if gethandcount(a)<gMAXCARDSINHAND then sendhand(tCOMBAT[a][i], "Combat", a, i)
            else senddeck(tCOMBAT[a][i], "Combat", a, i) end
          end
        end
      end
      if next(tCOMBAT[b])~=nil then
        r = true
        for i = #tCOMBAT[b], 1, -1 do
          if tCOMBAT[b][i].Class~="Insect" and tCOMBAT[b][i].Class~="Arachnid" and tCOMBAT[b][i].Habitat=="Home" then
            tCOMBAT[b][i].Wounds = 0
            if gethandcount(b)<gMAXCARDSINHAND then sendhand(tCOMBAT[b][i], "Combat", b, i)
            else senddeck(tCOMBAT[b][i], "Combat", b, i) end
          end
        end
      end
    elseif sp == "Yip! Yip!" then
      addpopup("Yappy Dog barks!")
      if next(tCOMBAT[b])~=nil then
        for i = 1, #tCOMBAT[b], 1 do
          if tCOMBAT[b][i].Tier<=4 then
            if math.random(100)<=50 or tCOMBAT[b][i].Special=="Fearful" then 
              tCOMBAT[b][i].Effect[4]=math.max(tCOMBAT[b][i].Effect[4],2)
              tCOMBAT[b][i].Effect[1]=math.max(tCOMBAT[b][i].Effect[1],2)
              addfloatytext(tCOMBAT[b][i], "-Fear-")
            else addfloatytext(tCOMBAT[b][i], "+Resist+") end
          end
        end
        r = true
      end
    else
    
    end
  end
  return r
end

function love.update()
  local i, j, k
  if gFADE > 0 then  -- fade in graphics
    gFADE = gFADE - gFADESPEED
    if gFADE < 0 then gFADE = 0 end
  end
  gFLASH = gFLASH + gFLASH_z  -- flash counter
  if gFLASH > 255 then  -- this affects everything that flashes (same speed)
    gFLASH = gFLASH_max
    gFLASH_z = 0 - gFLASH_z
  elseif gFLASH < gFLASH_min then
    gFLASH = gFLASH_min
    gFLASH_z = 0 - gFLASH_z
  else end
  if gPOPUP_COUNT > 0 then  -- Pop up message counter
    gPOPUP_COUNT = gPOPUP_COUNT - 1
    if gPOPUP_COUNT == 0 then gPOPUP_MSG = "" end
  end
  if gDRAGGING == true then gHOLDING = gHOLDING + 1 end  -- Dragging card
  if gGAMEMODE == "Battle" then  -- Card animations & animated text
    for i = 1, 2, 1 do   
      if next(tDECK[i]) ~= nil then  -- Card animations (deck)
        for j = 1, #tDECK[i], 1 do
          if tDECK[i][j].Animate == true then animatecard(tDECK[i][j]) end
          if next(tDECK[i][j].FloatyText) ~= nil then animatetext(tDECK[i][j]) end
        end
      end
      if next(tHAND[i]) ~= nil then  -- Card animations (hand)
        for j = 1, #tHAND[i], 1 do
          if tHAND[i][j].Animate == true then animatecard(tHAND[i][j]) end
          if next(tHAND[i][j].FloatyText) ~= nil then animatetext(tHAND[i][j]) end
        end
      end
      if next(tGRAVE[i]) ~= nil then  -- Card animations (graveyard)
        for j = 1, #tGRAVE[i], 1 do
          if tGRAVE[i][j].Animate == true then animatecard(tGRAVE[i][j]) end
          if next(tGRAVE[i][j].FloatyText) ~= nil then animatetext(tGRAVE[i][j]) end
        end
      end
      if next(tCOMBAT[i]) ~= nil then  -- Card animations (combat)
        for j = 1, #tCOMBAT[i], 1 do
          if tCOMBAT[i][j].Animate == true then animatecard(tCOMBAT[i][j]) end
          if next(tCOMBAT[i][j].FloatyText) ~= nil then animatetext(tCOMBAT[i][j]) end
        end
      end
      if next(tHERO[i]) ~= nil then  -- Card animations (hero)
        if tHERO[i].Animate == true then animatecard(tHERO[i]) end
        if next(tHERO[i].FloatyText) ~= nil then animatetext(tHERO[i]) end
      end
    end
  end
  if gENDING > 0 then 
    gENDING  = gENDING - 1
    if gENDING == 0 and gGAMEMODE == "Battle" then
      gENDING = gFPS * 5
      gFADE = 255
      gGAMEMODE = "Ending"
      endgame()
      --save cards / unlocked states
    elseif gENDING == 0 and gGAMEMODE == "Ending" then
      resetmainmenu()
      gGAMEMODE = "Main Menu"
    else end
  end
  if tCONTROL[gTURN] == "AI" and gENDING == 0 and gGAMEMODE == "Battle" and (AIloop<AImax_loops or love.timer.getTime()-AIlastmove>2) then AI() end -- Call the A.I. 
  if love.timer.getTime()-gTIMER<=(1/gFPS) and tCONTROL[gTURN]~="AI" then love.timer.sleep((1/gFPS)-(love.timer.getTime()-gTIMER)) end  -- frame rate
  gTIMER = love.timer.getTime()  -- new timer
end

function love.draw()
  love.graphics.scale(gRATIOX, gRATIOY)
  if gGAMEMODE == "Title" then drawtitlescreen()
  elseif gGAMEMODE == "Main Menu" then drawmainmenu()
  elseif gGAMEMODE == "Select Level" then drawselectlevel()
  elseif gGAMEMODE == "Deck Menu" then drawdeckmenu()
  elseif gGAMEMODE == "Deck Parade" then drawdeckparade(gDECK_SELECT)
  elseif gGAMEMODE == "File Menu" then drawfilemenu()
  elseif gGAMEMODE == "New Turn" then drawnewturn()
  elseif gGAMEMODE == "Battle" then drawbattle()
  elseif gGAMEMODE == "Ending" then drawending()
  else end
  drawpopuptext()
  if gFADE > 0 then 
    love.graphics.setColor(0, 0, 0, gFADE)
    love.graphics.rectangle("fill", 0, 0, 800, 600)
  else end
end

function drawpopuptext()
  local text, w, wrappedtext, n, h
  if gPOPUP_COUNT > 0 or gTYPING == true then
    if gTYPING==true then text=gTYPEDTEXT
    else text=gPOPUP_MSG end
    love.graphics.setFont(fHEADING)
    h = fHEADING:getHeight()
    w, wrappedtext = fHEADING:getWrap(text, 390)
    if next(wrappedtext)~=nil then n=#wrappedtext else n = 0 end
    love.graphics.setColor(0, 0, 0, 255)
    love.graphics.rectangle("fill", 400-(w/2)-5, 300-(n*h/2)-5, w+10, (n*h)+10)
    love.graphics.setColor(255, 255, 255, 255)
    love.graphics.rectangle("line", 400-(w/2)-5, 300-(n*h/2)-5, w+10, (n*h)+10)
    love.graphics.printf(text, 400-(w/2), 300-(n*h/2), w, "center")
  end
end

function drawtitlescreen()
  love.graphics.setFont(fTITLE)
  love.graphics.setColor(255, 215, 0, 255)
  love.graphics.printf("Habitat Wars", 200, 280, 400, "center")
end

function drawnewturn()
  love.graphics.setFont(fTITLE)
  love.graphics.setColor(255, 215, 0, 255)
  love.graphics.printf("P"..gTURN..": Your Turn", 200, 200, 400, "center")
  if gDAYNIGHT>=6 then
    love.graphics.printf("Daytime"..gDAYNIGHT-5, 200, 360, 400, "center")
  else
    love.graphics.setColor(50, 0, 200, 255)
    love.graphics.printf("Nighttime"..gDAYNIGHT, 200, 360, 400, "center")
  end
end

function drawmainmenu()
  local menu_text = {"Single Player", "Decks", "Pass and Play", "Exit"}
  local x, y = love.mouse.getX()/gRATIOX, love.mouse.getY()/gRATIOY
  local menu_x = {520, 420, 530, 400}
  local menu_y = {200, 275, 350, 425}
  local icon_y = {185, 260, 335, 410}
  local i, j, k
  drawtitle("Main Menu")
  drawcointotal()
  love.graphics.setFont(fHEADING)
  for i = 1, 4, 1 do
    love.graphics.setColor(255, 255, 255, 255)
    if x>=350 and x<=menu_x[i] and y>=menu_y[i] and y<=menu_y[i]+25 then love.graphics.setColor(255, 255, 255, gFLASH) end
    if x>=280 and x<=320 and y>=icon_y[i]+5 and y<=icon_y[i]+45 then love.graphics.setColor(255, 255, 255, gFLASH) end
    love.graphics.printf(menu_text[i], 350, menu_y[i], 200, "left")
    love.graphics.draw(iICON[i+4], 275, icon_y[i])
  end
  love.graphics.setColor(255, 255, 255, 255)
  love.graphics.printf("Sounds", 25, 500, 100, "center")
  love.graphics.printf("Music", 675, 500, 100, "center")
  if x>=55 and x<=95 and y>=530 and y<=570 then love.graphics.setColor(255, 255, 255, gFLASH)
  else love.graphics.setColor(255, 255, 255, 255) end
  if gSOUNDON == true then love.graphics.draw(iICON[4], 50, 525)
  else love.graphics.draw(iICON[3], 50, 525) end
  if x>=705 and x<=745 and y>=530 and y<=570 then love.graphics.setColor(255, 255, 255, gFLASH)
  else love.graphics.setColor(255, 255, 255, 255) end
  if gMUSICON == true then love.graphics.draw(iICON[2], 700, 525)
  else love.graphics.draw(iICON[1], 700, 525) end
end

function drawbattle()
  local i
  love.graphics.setColor(255, 255, 255, 255)
  for i = 1, 2, 1 do
    love.graphics.draw(iICON[17], xGRAVE[i]-30, yGRAVE[i]-32)
  end
  for i = 1, 2, 1 do
    checkhover(i)
  end
  for i = 1, 2, 1 do
    drawdeck(i)
    drawgrave(i)
    drawhand(i)
    drawcombat(i)
    drawhero(i)
  end
  for i = 1, 2, 1 do
    drawhover(i)
  end
  drawanimatedtext()
  drawgameinfo()
  drawselectedcard()
  drawheldcard()
end

function drawanimatedtext()
  love.graphics.setFont(fHEADING)
  local i, j, k
  local x, y, z, t
  for j = 1, 2, 1 do
    if next(tDECK[j])~=nil then
      for i = 1, #tDECK[j], 1 do
        if next(tDECK[j][i].FloatyText)~=nil then
          for k = 1, #tDECK[j][i].FloatyText, 1 do
            x = tDECK[j][i].FloatyText[k].X
            y = tDECK[j][i].FloatyText[k].Y-tDECK[j][i].FloatyText[k].Timer
            z = math.floor(255*((gFLOAT_TIMER-tDECK[j][i].FloatyText[k].Timer)/gFLOAT_TIMER))
            t = tDECK[j][i].FloatyText[k].Text
            love.graphics.setColor(0,0,0,z)
            love.graphics.printf(t, x-79, y+1, 160, "center")
            if string.find(t, "+") then love.graphics.setColor(0, 255, 0, z)
            elseif string.find(t, "-") then love.graphics.setColor(255, 0, 0, z)
            else love.graphics.setColor(255, 255, 255, z) end
            love.graphics.printf(t, x-100, y, 200, "center")
          end
        end
      end
    end
    if next(tHAND[j])~=nil then
      for i = 1, #tHAND[j], 1 do
        if next(tHAND[j][i].FloatyText)~=nil then
          for k = 1, #tHAND[j][i].FloatyText, 1 do
            x = tHAND[j][i].FloatyText[k].X
            y = tHAND[j][i].FloatyText[k].Y-tHAND[j][i].FloatyText[k].Timer
            z = math.floor(255*((gFLOAT_TIMER-tHAND[j][i].FloatyText[k].Timer)/gFLOAT_TIMER))
            t = tHAND[j][i].FloatyText[k].Text
            love.graphics.setColor(0,0,0,z)
            love.graphics.printf(t, x-79, y+1, 160, "center")
            if string.find(t, "+") then love.graphics.setColor(0, 255, 0, z)
            elseif string.find(t, "-") then love.graphics.setColor(255, 0, 0, z)
            else love.graphics.setColor(255, 255, 255, z) end
            love.graphics.printf(t, x-80, y, 160, "center")
          end
        end
      end
    end
    if next(tCOMBAT[j])~=nil then
      for i = 1, #tCOMBAT[j], 1 do
        if next(tCOMBAT[j][i].FloatyText)~=nil then
          for k = 1, #tCOMBAT[j][i].FloatyText, 1 do
            x = tCOMBAT[j][i].FloatyText[k].X
            y = tCOMBAT[j][i].FloatyText[k].Y-tCOMBAT[j][i].FloatyText[k].Timer
            z = math.floor(255*((gFLOAT_TIMER-tCOMBAT[j][i].FloatyText[k].Timer)/gFLOAT_TIMER))
            t = tCOMBAT[j][i].FloatyText[k].Text
            love.graphics.setColor(0,0,0,z)
            love.graphics.printf(t, x-79, y+1, 160, "center")
            if string.find(t, "+") then love.graphics.setColor(0, 255, 0, z)
            elseif string.find(t, "-") then love.graphics.setColor(255, 0, 0, z)
            else love.graphics.setColor(255, 255, 255, z) end
            love.graphics.printf(t, x-80, y, 160, "center")
          end
        end
      end
    end
    if next(tGRAVE[j])~=nil then
      for i = 1, #tGRAVE[j], 1 do
        if next(tGRAVE[j][i].FloatyText)~=nil then
          for k = 1, #tGRAVE[j][i].FloatyText, 1 do
            x = tGRAVE[j][i].FloatyText[k].X
            y = tGRAVE[j][i].FloatyText[k].Y-tGRAVE[j][i].FloatyText[k].Timer
            z = math.floor(255*((gFLOAT_TIMER-tGRAVE[j][i].FloatyText[k].Timer)/gFLOAT_TIMER))
            t = tGRAVE[j][i].FloatyText[k].Text
            love.graphics.setColor(0,0,0,z)
            love.graphics.printf(t, x-79, y+1, 160, "center")
            if string.find(t, "+") then love.graphics.setColor(0, 255, 0, z)
            elseif string.find(t, "-") then love.graphics.setColor(255, 0, 0, z)
            else love.graphics.setColor(255, 255, 255, z) end
            love.graphics.printf(t, x-80, y, 160, "center")
          end
        end
      end
    end
    if next(tHERO[j])~=nil then
      if next(tHERO[j].FloatyText)~=nil then
        for k = 1, #tHERO[j].FloatyText, 1 do
          x = tHERO[j].FloatyText[k].X
          y = tHERO[j].FloatyText[k].Y-tHERO[j].FloatyText[k].Timer
          z = math.floor(255*((gFLOAT_TIMER-tHERO[j].FloatyText[k].Timer)/gFLOAT_TIMER))
          t = tHERO[j].FloatyText[k].Text
          love.graphics.setColor(0,0,0,z)
          love.graphics.printf(t, x-79, y+1, 160, "center")
          if string.find(t, "+") then love.graphics.setColor(0, 255, 0, z)
          elseif string.find(t, "-") then love.graphics.setColor(255, 0, 0, z)
          else love.graphics.setColor(255, 255, 255, z) end
          love.graphics.printf(t, x-80, y, 160, "center")
        end
      end
    end
  end  
end

function checkhover(pl)
  local x, y = love.mouse.getX() / gRATIOX, love.mouse.getY() / gRATIOY
  local i, j, m, stack, width
  local hover = 0
  if next(tDECK[pl]) ~= nil then
    for i = 1, #tDECK[pl], 1 do
      tDECK[pl][i].Hover = false
    end
    if hover == 0 and (gZOOMED == false or gDRAGGING == true) then
      stack = math.min(#tDECK[pl], 5)
      if x>xDECK[pl]-(gCARDWIDTH/2)-12 and x<xDECK[pl]+(gCARDWIDTH/2)-12 and y>yDECK[pl]-(gCARDHEIGHT/2)-12 and y<yDECK[pl]+(gCARDHEIGHT/2)-12 and tDECK[pl][1].Clickable == true then
        tDECK[pl][1].Hover = true
        hover = 1
      end
    end
  end
  if next(tGRAVE[pl]) ~= nil then
    for i = 1, #tGRAVE[pl], 1 do
      tGRAVE[pl][i].Hover = false
    end
    if hover == 0 and (gZOOMED == false or gDRAGGING == true) then
      i = #tGRAVE[pl]
      if x>xGRAVE[pl]-(gCARDWIDTH/2) and x<xGRAVE[pl]+(gCARDWIDTH/2) and y>yGRAVE[pl]-(gCARDHEIGHT/2) and y<yGRAVE[pl]+(gCARDHEIGHT/2) and tGRAVE[pl][i].Clickable == true then
        tGRAVE[pl][i].Hover = true
        hover = i
      end
    end
  end
  if next(tHAND[pl]) ~= nil then
    for i = 1, #tHAND[pl], 1 do
      tHAND[pl][i].Hover = false
    end
    if hover == 0 and (gZOOMED == false or gDRAGGING == true) then
      m = (#tHAND[pl]+1)/2
      width = gHANDWIDTH-gCARDWIDTH
      for i = 1, #tHAND[pl], 1 do
        j = (i-m)/m
        if x>xHAND[pl]+(j*width)-(gCARDWIDTH/2) and x<xHAND[pl]+(j*width)+(gCARDWIDTH/2) and y>yHAND[pl]-(gHANDHEIGHT/2) and y<yHAND[pl]+(gHANDHEIGHT/2) and tHAND[pl][i].Clickable == true then
          hover = i
        end
      end
      if hover ~= 0 then tHAND[pl][hover].Hover=true end
    end
  end
  if next(tCOMBAT[pl]) ~= nil then
    for i = 1, #tCOMBAT[pl], 1 do
      tCOMBAT[pl][i].Hover = false
    end
    if hover == 0 and (gZOOMED == false or gDRAGGING == true) then
      m = (#tCOMBAT[pl]+1)/2
      width = gCOMBATWIDTH-gCARDWIDTH
      for i = 1, #tCOMBAT[pl], 1 do
        j=(i-m)/m
        if x>xCOMBAT[pl]+(j*width)-(gCARDWIDTH/2) and x<xCOMBAT[pl]+(j*width)+(gCARDWIDTH/2) and y>yCOMBAT[pl]-(gCARDHEIGHT/2) and y<yCOMBAT[pl]+(gCARDHEIGHT/2) and tCOMBAT[pl][i].Clickable == true then 
          hover = i 
        end
      end
      if hover ~= 0 then tCOMBAT[pl][hover].Hover=true end
    end
  end
  if next(tHERO[pl]) ~= nil then
    tHERO[pl].Hover = false
    if hover == 0 and (gZOOMED == false or gDRAGGING == true) then
      if x>xHERO[pl]-(gCARDWIDTH/2) and x<xHERO[pl]+(gCARDWIDTH/2) and y>yHERO[pl]-(gCARDHEIGHT/2) and y<yHERO[pl]+(gCARDHEIGHT/2) and tHERO[pl].Clickable == true then
        tHERO[pl].Hover = true
        hover = 1
      end
    end
  end
end

function drawdeck(pl)
  local i, stack
  if next(tDECK[pl]) ~= nil then
    stack = math.min(#tDECK[pl], 5)
    for i = stack, 1, -1 do
      if tDECK[pl][i].Hover == false then drawcard(tDECK[pl][i], xDECK[pl]-((5-i)*3), yDECK[pl]-((5-i)*3), gCARDWIDTH, gCARDHEIGHT) end 
    end
  end
end

function drawgrave(pl)
  local i
  if next(tGRAVE[pl]) ~= nil then
    for i = 1, #tGRAVE[pl], 1 do
      if tGRAVE[pl][i].Hover == false then drawcard(tGRAVE[pl][i], tGRAVE[pl][i].X, tGRAVE[pl][i].Y, gCARDWIDTH, gCARDHEIGHT, tGRAVE[pl][i].Angle) end 
    end
  end
end

function arrangehand(pl)
  local width, height = gHANDWIDTH-gCARDWIDTH, gHANDHEIGHT-gCARDHEIGHT
  local i, j, m
  if next(tHAND[pl])~=nil then
    m = (#tHAND[pl]+1)/2
    for i = 1, #tHAND[pl], 1 do
      j = (i-m)/m
      tHAND[pl][i].OriginalX = tHAND[pl][i].X
      tHAND[pl][i].OriginalY = tHAND[pl][i].Y
      tHAND[pl][i].TargetX = xHAND[pl]+(j*width)
      tHAND[pl][i].TargetY = yHAND[pl]+height-(math.cos(j)*height)
      tHAND[pl][i].Angle = j*gHANDSLANT
      tHAND[pl][i].Animate = true
      tHAND[pl][i].AniCount = gANIMATION_TIME
    end
  end
end

function drawhand(pl)
  local i
  if next(tHAND[pl]) ~= nil then
    for i = 1, #tHAND[pl], 1 do
      if tHAND[pl][i].Hover == false then
        drawcard(tHAND[pl][i], tHAND[pl][i].X, tHAND[pl][i].Y, gCARDWIDTH, gCARDHEIGHT, tHAND[pl][i].Angle)
      end
    end
  end
end

function arrangecombat(pl)
  local width = gCOMBATWIDTH-gCARDWIDTH
  local i, j, m
  if next(tCOMBAT[pl])~=nil then
    m = (#tCOMBAT[pl]+1) / 2
    for i = 1, #tCOMBAT[pl], 1 do
      j = (i-m)/m
      tCOMBAT[pl][i].OriginalX = tCOMBAT[pl][i].X
      tCOMBAT[pl][i].OriginalY = tCOMBAT[pl][i].Y
      tCOMBAT[pl][i].TargetX = xCOMBAT[pl]+(j*width)
      tCOMBAT[pl][i].TargetY = yCOMBAT[pl]
      tCOMBAT[pl][i].Angle = 0
      tCOMBAT[pl][i].Animate = true
      tCOMBAT[pl][i].AniCount = gANIMATION_TIME
    end
  end
end

function drawcombat(pl)
  local i
  if next(tCOMBAT[pl]) ~= nil then
    for i = 1, #tCOMBAT[pl], 1 do
      if tCOMBAT[pl][i].Hover == false then
        drawcard(tCOMBAT[pl][i], tCOMBAT[pl][i].X, tCOMBAT[pl][i].Y, gCARDWIDTH, gCARDHEIGHT, tCOMBAT[pl][i].Angle)
      end
    end
  end
end

function drawhero(pl)
  if next(tHERO[pl])~=nil then
    if tHERO[pl].Hover == false then
      drawcard(tHERO[pl], tHERO[pl].X, tHERO[pl].Y, gCARDWIDTH, gCARDHEIGHT, tHERO[pl].Angle)
    end
  end
end

function drawcard(card, x, y, width, height, angle, alpha)
  angle = angle or 0
  alpha = alpha or 255
  local h = gethealth(card)
  local c = card.Tier
  local a = getpower(card)
  local p = getprice(card)
  local text = ""
  local sx, sy = width/300, height/420
  local i, j
  if c == 0 and card.Type == "Hero" then c = "H" end
  angle = math.rad(angle)
  if gGAMEMODE == "Battle" and card.Used == true and width<=gCARDWIDTH then alpha = 100 end
  if card.FaceUp == true then
    love.graphics.setColor(255, 255, 255, alpha)
    love.graphics.draw(card.Front, x, y, angle, sx, sy, 150, 210)
    love.graphics.setColor(255, 255, 255, alpha)
    love.graphics.draw(card.Image, x, y, angle, sx, sy, 80, 155)
    love.graphics.draw(iSQUARE, x, y, angle, sx, sy, 80, 155)
    if a > 0 or card.Type == "Ally" then love.graphics.draw(iICON[10], x, y, angle, sx, sy, -78, -132) end
    love.graphics.draw(iICON[9], x, y, angle, sx, sy, 144, 204)
    love.graphics.draw(iICON[12], x, y, angle, sx, sy, -84, 204) 
    love.graphics.setFont(fTITLE)
    love.graphics.setColor(0, 0, 0, alpha)
    if a > 0 or card.Type == "Ally" then love.graphics.printf(a, x, y, 68, "center", angle, sx, sy, -78, -156) end
    love.graphics.printf(c, x, y, 60, "center", angle, sx, sy, 144, 186)
    if p<10 and gGAMEMODE=="Deck Parade" then love.graphics.printf("$"..p, x, y, 60, "center", angle, sx, sy, -84, 190) end
    if card.Type=="Hero" and gGAMEMODE~="Battle" then
      love.graphics.printf("Hero", x, y, 160, "center", angle-math.rad(90), sx, sy, -2, 133)
      love.graphics.printf("Card", x, y, 160, "center", angle+math.rad(90), sx, sy, 158, 137)
    end
    love.graphics.setColor(255, 255, 255, alpha)
    if a > 0 or card.Type == "Ally" then love.graphics.printf(a, x, y, 68, "center", angle, sx, sy, -80, -158) end
    love.graphics.printf(c, x, y, 60, "center", angle, sx, sy, 146, 188)
    if p<10 and gGAMEMODE=="Deck Parade" then love.graphics.printf("$"..p, x, y, 60, "center", angle, sx, sy, -82, 192) end
    if card.Type=="Hero" and gGAMEMODE~="Battle" then
      love.graphics.printf("Hero", x, y, 160, "center", angle-math.rad(90), sx, sy, 0, 135)
      love.graphics.printf("Card", x, y, 160, "center", angle+math.rad(90), sx, sy, 160, 135)
    end
    love.graphics.setFont(fHEADING)
    love.graphics.setColor(0, 0, 0, alpha)
    if string.len(card.Name)<=12 then love.graphics.printf(card.Name, x, y, 170, "left", angle, sx, sy, 76, 200) end
    love.graphics.printf(card.Class, x, y, 160, "center", angle, sx, sy, 78, -150)
    if p>=10 and gGAMEMODE=="Deck Parade" then love.graphics.printf("$"..p, x, y, 60, "center", angle, sx, sy, -84, 186) end
    love.graphics.setColor(255, 255, 255, alpha)
    if string.len(card.Name)<=12 then love.graphics.printf(card.Name, x, y, 170, "left", angle, sx, sy, 78, 202) end
    love.graphics.printf(card.Class, x, y, 160, "center", angle, sx, sy, 80, -148)
    if p>=10 and gGAMEMODE=="Deck Parade" then love.graphics.printf("$"..p, x, y, 60, "center", angle, sx, sy, -82, 188) end
    if card.MaxHealth > 0 then
      if h > card.MaxHealth*0.7 then love.graphics.setColor(0, 255, 0, alpha) 
      elseif h > card.MaxHealth*0.4 then love.graphics.setColor(255, 150, 0, alpha)
      else love.graphics.setColor(255, 0, 0, alpha) end
      love.graphics.draw(iICON[11], x, y, angle, sx, sy, 146, -146) 
      love.graphics.setColor(0, 0, 0, alpha)
      love.graphics.printf(h, x, y, 40, "center", angle, sx, sy, 149, -148)
      love.graphics.printf("/", x, y, 60, "center", angle, sx, sy, 143, -158)
      love.graphics.printf(card.MaxHealth, x, y, 40, "center", angle, sx, sy, 119, -172)
      love.graphics.setColor(255, 255, 255, alpha)
      love.graphics.printf(h, x, y, 40, "center", angle, sx, sy, 151, -146)
      love.graphics.printf("/", x, y, 60, "center", angle, sx, sy, 144, -157)
      love.graphics.printf(card.MaxHealth, x, y, 40, "center", angle, sx, sy, 121, -170)
    end 
    if card.Special ~= "None" then 
      text = card.Special..": "..card.SpDesc
      if card.SpVal ~= 0 then text = text.." Costs "..card.SpVal.."."
      elseif card.Type ~= "Action" then text = text.." Automatic."
      else end
    end
    love.graphics.setFont(fDEFAULT)
    love.graphics.setColor(0, 0, 0, alpha)
    if string.len(card.Name)>=13 then love.graphics.printf(card.Name, x, y, 170, "left", angle, sx, sy, 76, 200) end
    love.graphics.printf(card.Tag, x, y, 270, "left", angle, sx, sy, 134, -13)
    if text ~= "" then love.graphics.printf(text, x, y, 270, "left", angle, sx, sy, 134, -81) end
    love.graphics.printf(card.Habitat, x, y, 185, "left", angle, sx, sy, 76, 174)
    love.graphics.setColor(255, 255, 255, alpha)
    if string.len(card.Name)>=13 then love.graphics.printf(card.Name, x, y, 170, "left", angle, sx, sy, 78, 202) end
    love.graphics.printf(card.Tag, x, y, 270, "left", angle, sx, sy, 135, -12)
    if text ~= "" then love.graphics.printf(text, x, y, 270, "left", angle, sx, sy, 135, -80) end
    love.graphics.printf(card.Habitat, x, y, 185, "left", angle, sx, sy, 78, 176)
    j = 0
    for i = 1, #tEFFECTS, 1 do
      if card.Effect[i]>0 then
        j = j + 1
        love.graphics.setFont(fDEFAULT)
        if j<=4 then
          love.graphics.setColor(0, 0, 0, alpha)
          love.graphics.printf(tEFFECTS[i], x, y, 70, "center", angle, sx, sy, 148, 175-(j*35))
          love.graphics.setColor(255, 255, 255, alpha)
          love.graphics.printf(tEFFECTS[i], x, y, 70, "center", angle, sx, sy, 150, 177-(j*35))
        elseif j>4 and j<=8 then
          love.graphics.setColor(0, 0, 0, alpha)
          love.graphics.printf(tEFFECTS[i], x, y, 70, "center", angle, sx, sy, -80, 175-((j-4)*35))
          love.graphics.setColor(255, 255, 255, alpha)
          love.graphics.printf(tEFFECTS[i], x, y, 70, "center", angle, sx, sy, -78, 177-((j-4)*35))
        else end
        love.graphics.setFont(fTINY)
        if j<=4 then
          love.graphics.setColor(0, 0, 0, alpha)
          love.graphics.printf(tRESULTS[i], x, y, 70, "center", angle, sx, sy, 149, 160-(j*35))
          love.graphics.setColor(255, 255, 255, alpha)
          love.graphics.printf(tRESULTS[i], x, y, 70, "center", angle, sx, sy, 150, 161-(j*35))
        elseif j>4 and j<=8 then
          love.graphics.setColor(0, 0, 0, alpha)
          love.graphics.printf(tRESULTS[i], x, y, 70, "center", angle, sx, sy, -80, 160-((j-4)*35))
          love.graphics.setColor(255, 255, 255, alpha)
          love.graphics.printf(tRESULTS[i], x, y, 70, "center", angle, sx, sy, -79, 161-((j-4)*35))
        else end
      end
    end
  else
    love.graphics.setColor(255, 255, 255, 255)
    love.graphics.draw(card.Back, x, y, angle, sx, sy, 150, 210)
  end
end

function drawhover(pl)
  local i
  if next(tDECK[pl]) ~= nil then
    if tDECK[pl][1].Hover == true then
      if pl==1 then
        drawcard(tDECK[pl][1], xDECK[pl]+(gCARDWIDTH*0.2)-12, yDECK[pl]-(gCARDHEIGHT*0.2)-12, gCARDWIDTH*1.4, gCARDHEIGHT*1.4)
      else
        drawcard(tDECK[pl][1], xDECK[pl]+(gCARDWIDTH*0.2)-12, yDECK[pl]+(gCARDHEIGHT*0.2)-12, gCARDWIDTH*1.4, gCARDHEIGHT*1.4)
      end
    end
  end
  if next(tGRAVE[pl]) ~= nil then
    i = #tGRAVE[pl]
    if tGRAVE[pl][i].Hover == true then
      if pl==1 then
        drawcard(tGRAVE[pl][i], xGRAVE[pl]-(gCARDWIDTH*0.2), yGRAVE[pl]-(gCARDHEIGHT*0.2), gCARDWIDTH*1.4, gCARDHEIGHT*1.4, tGRAVE[pl][i].Angle)
      else
        drawcard(tGRAVE[pl][i], xGRAVE[pl]-(gCARDWIDTH*0.2), yGRAVE[pl]+(gCARDHEIGHT*0.2), gCARDWIDTH*1.4, gCARDHEIGHT*1.4, tGRAVE[pl][i].Angle)
      end
    end
  end
  if next(tHAND[pl]) ~= nil then
    for i = 1, #tHAND[pl], 1 do
      if tHAND[pl][i].Hover == true then
        if pl == 1 then drawcard(tHAND[pl][i], tHAND[pl][i].X, tHAND[pl][i].Y-(gCARDHEIGHT*0.2), gCARDWIDTH*1.4, gCARDHEIGHT*1.4, tHAND[pl][i].Angle)
        else drawcard(tHAND[pl][i], tHAND[pl][i].X, tHAND[pl][i].Y+(gCARDHEIGHT*0.2), gCARDWIDTH*1.4, gCARDHEIGHT*1.4, tHAND[pl][i].Angle) end
      end
    end
  end
  if next(tCOMBAT[pl]) ~= nil then
    for i = 1, #tCOMBAT[pl], 1 do
      if tCOMBAT[pl][i].Hover == true then drawcard(tCOMBAT[pl][i], tCOMBAT[pl][i].X, tCOMBAT[pl][i].Y, gCARDWIDTH*1.4, gCARDHEIGHT*1.4, tCOMBAT[pl][i].Angle) end
    end
  end
  if next(tHERO[pl]) ~= nil then
    if tHERO[pl].Hover == true then
      drawcard(tHERO[pl], tHERO[pl].X, tHERO[pl].Y, gCARDWIDTH*1.4, gCARDHEIGHT*1.4, tHERO[pl].Angle)
    end
  end
end

function drawtitle(title)
  love.graphics.setFont(fTITLE)
  love.graphics.setColor(255, 215, 0, 255)
  love.graphics.printf(title, 200, 15, 400, "center")
end

function drawmainmenubutton()
  local x, y = love.mouse.getX() / gRATIOX, love.mouse.getY() / gRATIOY
  love.graphics.setFont(fHEADING)
  if x>=0 and x<=180 and y>=0 and y<=50 then love.graphics.setColor(255, 255, 255, gFLASH)
  else love.graphics.setColor(255, 255, 255, 255) end
  love.graphics.printf("Main Menu", 50, 20, 130, "left")
  love.graphics.draw(iICON[15], 0, 0)
end

function drawdeckmenubutton()
  local x, y = love.mouse.getX() / gRATIOX, love.mouse.getY() / gRATIOY
  love.graphics.setFont(fHEADING)
  if x>=0 and x<=180 and y>=0 and y<=50 then love.graphics.setColor(255, 255, 255, gFLASH)
  else love.graphics.setColor(255, 255, 255, 255) end
  love.graphics.printf("Deck Menu", 50, 20, 130, "left")
  love.graphics.draw(iICON[15], 0, 0)
end

function drawcointotal()
  love.graphics.setFont(fHEADING)
  love.graphics.setColor(255, 255, 255, 255)
  love.graphics.draw(iICON[12], 755, 15, 0, 0.5, 0.5)
  love.graphics.printf("$"..gCOINS, 600, 20, 150, "right")
end

function drawselectlevel()
  local x, y = love.mouse.getX() / gRATIOX, love.mouse.getY() / gRATIOY
  local i, j, start, finish, cards, unlocked
  local width = 150
  local height = 210
  drawtitle("Choose a Level")
  drawcointotal()
  drawmainmenubutton()
  if next(tHABITAT) ~= nil then
    start = math.max(gLEVEL_CURSOR-2, 1)
    finish = math.min(gLEVEL_CURSOR+2, #tHABITAT)
    love.graphics.setFont(fHEADING)
    for i = start, finish, 1 do
      j = i-gLEVEL_CURSOR
      if tHABITAT[i].Unlock > 0 then love.graphics.setColor(255, 255, 255, 255)
      else love.graphics.setColor(80, 80, 80, 255) end
      love.graphics.draw(tHABITAT[i].Image, 400+(j*(width+10)), 290, 0, width/300, height/420, 150, 210)
      love.graphics.setColor(0, 0, 0, 255)
      love.graphics.printf(tHABITAT[i].Name, 326+(j*(width+10)), 281, 150, "center")
      if tHABITAT[i].Unlock > 0 then love.graphics.setColor(255, 255, 255, 255)
      else love.graphics.setColor(80, 80, 80, 255) end
      love.graphics.printf(tHABITAT[i].Name, 324+(j*(width+10)), 279, 150, "center")
      if tHABITAT[i].Unlock >= 1 then 
        if x>=325+(j*(width+10)) and x<=475+(j*(width+10)) and y>=415 and y<=445 then love.graphics.setColor(0, 255, 0, gFLASH)
        else love.graphics.setColor(0, 255, 0, 255) end
        love.graphics.printf("EASY $"..tHABITAT[i].Easy, 325+(j*(width+10)), 415, 150, "center")
      else
        love.graphics.setColor(80, 80, 80, 255)
        love.graphics.printf("LOCKED", 325+(j*(width+10)), 415, 150, "center")
      end
      if tHABITAT[i].Unlock >= 2 then
        if x>=325+(j*(width+10)) and x<=475+(j*(width+10)) and y>=445 and y<=475 then love.graphics.setColor(255, 150, 0, gFLASH)
        else love.graphics.setColor(255, 150, 0, 255) end
        love.graphics.printf("MEDIUM $"..tHABITAT[i].Medium, 325+(j*(width+10)), 445, 150, "center")
      else
        love.graphics.setColor(80, 80, 80, 255)
        love.graphics.printf("LOCKED", 325+(j*(width+10)), 445, 150, "center")
      end
      if tHABITAT[i].Unlock >= 3 then 
        if x>=325+(j*(width+10)) and x<=475+(j*(width+10)) and y>=475 and y<=505 then love.graphics.setColor(255, 0, 0, gFLASH)
        else love.graphics.setColor(255, 0, 0, 255) end
        love.graphics.printf("HARD $"..tHABITAT[i].Hard, 325+(j*(width+10)), 475, 150, "center")
      else
        love.graphics.setColor(80, 80, 80, 255)
        love.graphics.printf("LOCKED", 325+(j*(width+10)), 475, 150, "center")
      end
      cards, unlocked = getnumunlockedcards(tHABITAT[i].Name)
      if cards > 0 and tHABITAT[i].Unlock >= 1 then
        love.graphics.setColor(255, 255, 255, 255)
        love.graphics.printf("Unlocked "..unlocked.."/"..cards, 325+(j*(width+10)), 515, 150, "center")
      end
    end
    if x>=25 and x<=75 and y>=275 and y<=325 then love.graphics.setColor(255, 255, 255, gFLASH)
    else love.graphics.setColor(255, 255, 255, 255) end
    if gLEVEL_CURSOR > 1 then love.graphics.draw(iICON[13], 25, 275) end
    if x>=725 and x<=775 and y>=275 and y<=325 then love.graphics.setColor(255, 255, 255, gFLASH)
    else love.graphics.setColor(255, 255, 255, 255) end
    if gLEVEL_CURSOR < #tHABITAT then love.graphics.draw(iICON[14], 725, 275) end
    if x>775 and gLEVEL_CURSOR < #tHABITAT then
      gDELAYACTION = gDELAYACTION + 1
      if gDELAYACTION > (gFPS*0.33) then
        gLEVEL_CURSOR = gLEVEL_CURSOR + 1
        gDELAYACTION = 0
      end
    elseif x<25 and gLEVEL_CURSOR > 1 then
      gDELAYACTION = gDELAYACTION + 1
      if gDELAYACTION > (gFPS*0.33) then
        gLEVEL_CURSOR = gLEVEL_CURSOR - 1
        gDELAYACTION = 0
      end
    else gDELAYACTION = 0 end
  end
end

function drawdeckmenu()
  local x, y = love.mouse.getX() / gRATIOX, love.mouse.getY() / gRATIOY
  local a, b
  local cards, unlocked
  local i, j
  drawtitle("P"..gTURN..": Choose a Deck")
  drawcointotal()
  drawmainmenubutton()
  a, b, i = 10, 60, 0
  if next(tHABITAT)~=nil then
    repeat
      i = i + 1
      if tHABITAT[i].Unlock > 0 then 
        if x>a and x<a+120 and y>b and y<b+120 then love.graphics.setColor(255, 255, 255, gFLASH)
        else love.graphics.setColor(180, 180, 180, 255) end
      else love.graphics.setColor(80, 80, 80, 255) end
      love.graphics.draw(tHABITAT[i].Square, a, b)
      love.graphics.setFont(fHEADING)
      if tHABITAT[i].Unlock > 0 then
        love.graphics.setColor(0, 0, 0, 255)
        love.graphics.printf(tHABITAT[i].Name, a, b+50, 120, "center")
        love.graphics.setColor(255, 255, 255, 255)
        love.graphics.printf(tHABITAT[i].Name, a-2, b+48, 120, "center")
        love.graphics.setFont(fDEFAULT)
        cards, unlocked = getnumunlockedcards(tHABITAT[i].Name)
        love.graphics.printf("Unlocked "..unlocked.."/"..cards, a, b+125, 120, "center")
      else
        love.graphics.setColor(255, 255, 255, 255)
        love.graphics.draw(iICON[16], a+40, b+40, 0, 0.25, 0.25, 0, 0)
      end
      a = a + 140
      if a > 630 then
        a = 10
        b = b + 145
      end
    until i==#tHABITAT or i==15
  end
  j = i
  i = 0
  love.graphics.setFont(fHEADING)
  if next(tSAVES)~=nil and j<15 then
    repeat
      i = i + 1
      if x>a and x<a+120 and y>b and y<b+120 then love.graphics.setColor(255, 255, 255, gFLASH)
      else love.graphics.setColor(180, 180, 180, 255) end
      love.graphics.draw(iTILE, a, b)
      love.graphics.setColor(0, 0, 0, 255)
      love.graphics.printf(tSAVES[i], a, b+50, 120, "center")
      love.graphics.setColor(255, 255, 255, 255)
      love.graphics.printf(tSAVES[i], a-2, b+48, 120, "center")
      a = a + 140
      if a > 630 then
        a = 10
        b = b + 145
      end
    until j+i==15 or #tSAVES==i
  end  
  love.graphics.setFont(fHEADING)
  if getdeckcount(gTURN)>0 then
    if x>700 and x<790 and y>70 and y<120 then love.graphics.setColor(255, 255, 255, gFLASH)
    else love.graphics.setColor(180, 180, 180, 255) end
  else love.graphics.setColor(80, 80, 80, 255) end
  love.graphics.rectangle("line", 700, 65, 90, 60)
  love.graphics.printf("Blank Deck", 700, 70, 90, "center")
  if next(tSAVES) ~= nil then
    if x>700 and x<790 and y>135 and y<185 then love.graphics.setColor(255, 255, 255, gFLASH)
    else love.graphics.setColor(180, 180, 180, 255) end
  else love.graphics.setColor(80, 80, 80, 255) end
  love.graphics.rectangle("line", 700, 130, 90, 60)
  love.graphics.printf("Load Deck", 700, 135, 90, "center")
  if getdeckcount(gTURN)==gDECKSIZE and next(tHERO[gTURN])~=nil then
    if x>700 and x<790 and y>200 and y<250 then love.graphics.setColor(255, 255, 255, gFLASH)
    else love.graphics.setColor(180, 180, 180, 255) end
  else love.graphics.setColor(80, 80, 80, 255) end
  love.graphics.rectangle("line", 700, 195, 90, 60)
  love.graphics.printf("Save Deck", 700, 200, 90, "center")
  if x>700 and x<790 and y>265 and y<315 then love.graphics.setColor(255, 255, 255, gFLASH)
  else love.graphics.setColor(180, 180, 180, 255) end
  love.graphics.rectangle("line", 700, 260, 90, 60)
  love.graphics.printf("Auto Deck", 700, 265, 90, "center")
  if next(tSAVES) ~= nil then
    if x>700 and x<790 and y>330 and y<380 then love.graphics.setColor(255, 255, 255, gFLASH)
    else love.graphics.setColor(180, 180, 180, 255) end
  else love.graphics.setColor(80, 80, 80, 255) end
  love.graphics.rectangle("line", 700, 325, 90, 60)
  love.graphics.printf("Erase Deck", 700, 330, 90, "center")
  if getdeckcount(gTURN)==gDECKSIZE and next(tHERO[gTURN])~=nil then
    if x>700 and x<790 and y>480 and y<530 then love.graphics.setColor(0, 255, 0, gFLASH)
    else love.graphics.setColor(0, 255, 0, 255) end
    love.graphics.rectangle("line", 700, 475, 90, 60)
    if gTURN==1 and tCONTROL[2]=="Player" then
      love.graphics.draw(iICON[14], 740, 480)
      love.graphics.printf("P2", 700, 491, 40, "center")
    else
      love.graphics.printf("Done", 700, 491, 90, "center")
    end
  end
  drawcloseup()
  drawminideck(gTURN)
  drawheldcard()
end

function drawminideck(pl)
  local x, y = love.mouse.getX() / gRATIOX, love.mouse.getY() / gRATIOY
  local hover = 0
  local i
  if next(tHERO[pl])~=nil then
    if x>=0 and x<=40 and y>=540 and y<=600 then hover = -1
    else drawcard(tHERO[pl], 20, 570, 40, 60) end
  end
  if next(tDECK[pl])~=nil then
    for i = 1, #tDECK[pl], 1 do
      if x>=((i-1)*19)+40 and x<=((i-1)*19)+80 and y>=540 and y<=600 then hover = i end
    end
    for i = 1, #tDECK[pl], 1 do
      if i~=hover then drawcard(tDECK[pl][i], ((i-1)*19)+60, 570, 40, 60) end
    end
  end
  if hover == -1 then drawcard(tHERO[pl], 50, 530, 100, 140)
  elseif hover ~=0 then drawcard(tDECK[pl][hover], math.min(math.max(((hover-1)*19)+60,50),750), 530, 100, 140) 
  else end
end

function drawdeckparade()
  local x, y = love.mouse.getX() / gRATIOX, love.mouse.getY() / gRATIOY
  local a, b, i, hover, hover_x, hover_y
  drawtitle("Deck: "..tHABITAT[gDECK_SELECT].Name)
  drawcointotal()
  drawdeckmenubutton()
  if next(tLIST)~=nil then
    a, b, i, hover = 0, 60, 0, 0
    repeat
      i = i + 1
      if x>a and x<a+110 and y>b and y<b+154 and gZOOMED == false then
        hover = i
        hover_x = a
        hover_y = b
      end
      drawcard(tLIST[i], a+55, b+77, 110, 154)
      if tLIST[i].Unlocked=="Locked" then
        love.graphics.setColor(255, 255, 255, 200)
        love.graphics.draw(iBLANK, a+55, b+77, 0, 110/300, 154/420, 150, 210)
        love.graphics.setColor(255, 255, 255, 255)
        love.graphics.draw(iICON[16], a+55, b+77, 0, 110/300, 154/420, 80, 155)
        love.graphics.setFont(fDEFAULT)
        love.graphics.printf("LOCKED", a, b+90, 110, "center")
      end
      a = a + 115
      if a>690 then
        a = 0
        b = b + 160
      end
    until i==#tLIST or i==21
    if hover~=0 and tLIST[hover].Unlocked=="Unlocked" then
      love.graphics.setColor(255, 255, 255, 200)
      love.graphics.draw(iBLANK, hover_x+55, hover_y+77, 0, 110/300, 154/420, 150, 210)
      love.graphics.setFont(fDEFAULT)
      love.graphics.setColor(255, 255, 255, 255)
      love.graphics.printf("Owned:", hover_x, hover_y+65, 60, "right")
      love.graphics.printf(tLIST[hover].Copies.."/"..tLIST[hover].Maximum, hover_x+65, hover_y+65, 65, "left")
      love.graphics.printf("In Deck:", hover_x, hover_y+85, 60, "right")
      love.graphics.printf(getcountindeckbyid(tLIST[hover].ID,gTURN).."/"..tLIST[hover].Copies, hover_x+65, hover_y+85, 65, "left")
    end
  end
  drawcloseup()
  drawminideck(gTURN)
  drawheldcard()
end

function drawcloseup()
  if gZOOMED == true and gDRAGGING == false then
    love.graphics.setColor(0, 0, 0, 200)
    love.graphics.rectangle("fill", 0, 60, 800, 480)
    drawcard(tCLICK[1].Card, 400, 270, 300, 420)
    if tCLICK[1].Card.Unlocked=="Unlocked" then
      if tCLICK[1].Source=="List" then drawgenericbutton(550, 60, 150, 80) end
      drawgenericbutton(250, 480, 150, 54)
      drawgenericbutton(400, 480, 150, 54)
      love.graphics.setFont(fDEFAULT)
      love.graphics.setColor(255, 255, 255, 255)
      love.graphics.printf("You own "..tCLICK[1].Card.Copies.."/"..tCLICK[1].Card.Maximum.." copies of this card", 250, 485, 140, "center")
      love.graphics.printf("You have "..getcountindeckbyid(tCLICK[1].Card.ID,gTURN).." of these cards in your current deck", 400, 485, 140, "center")
      if getcountindeckbyid(tCLICK[1].Card.ID,gTURN)<tCLICK[1].Card.Copies and tCLICK[1].Source=="List" then
        drawgenericbutton(550, 260, 150, 80)
        love.graphics.setFont(fTITLE)
        love.graphics.setColor(255, 215, 0, 255)
        love.graphics.printf("+", 555, 265, 140, "center")
        love.graphics.setFont(fHEADING)
        love.graphics.setColor(255, 255, 255, 255)
        love.graphics.printf("Add to deck", 555, 300, 140, "center")
      end
      if getcountindeckbyid(tCLICK[1].Card.ID,gTURN)>0 then
        drawgenericbutton(100, 260, 150, 80)
        love.graphics.setFont(fTITLE)
        love.graphics.setColor(255, 215, 0, 255)
        love.graphics.printf("-", 105, 265, 140, "center")
        love.graphics.setFont(fHEADING)
        love.graphics.setColor(255, 255, 255, 255)
        love.graphics.printf("Remove", 105, 300, 140, "center")
      end
      if gCOINS>=tCLICK[1].Card.Price and tCLICK[1].Card.Copies<tCLICK[1].Card.Maximum and tCLICK[1].Source=="List" then
        love.graphics.setFont(fHEADING)
        love.graphics.setColor(255, 255, 255, 255)
        love.graphics.printf("BUY CARD", 555, 65, 140, "center")
        love.graphics.setFont(fTITLE)
        love.graphics.setColor(255, 215, 0, 255)
        love.graphics.printf("$"..tCLICK[1].Card.Price, 555, 95, 140, "center")
      elseif tCLICK[1].Card.Copies>=tCLICK[1].Card.Maximum and tCLICK[1].Source=="List" then
        love.graphics.setFont(fHEADING)
        love.graphics.setColor(255, 0, 0, 255)
        love.graphics.printf("Sold out", 555, 88, 140, "center")
      elseif gCOINS<tCLICK[1].Card.Price and tCLICK[1].Source=="List" then
        love.graphics.setFont(fHEADING)
        love.graphics.setColor(255, 0, 0, 255)
        love.graphics.printf("Out of price range", 555, 72, 140, "center")
      else end
    end
  end
end

function drawselectedcard()
  local x, y = love.mouse.getX() / gRATIOX, love.mouse.getY() / gRATIOY
  local moves
  if gZOOMED == true and gDRAGGING == false then
    moves = calculatemoves(tCLICK[1].Player)
    drawcard(tCLICK[1].Card, 400, 300, 300, 420)
    love.graphics.setFont(fHEADING)
    if moves[1]==true then      -- March
      drawgenericbutton(550, 90, 150, 84)
      if x>550 and x<700 and y>90 and y<174 then love.graphics.setColor(255, 255, 255, 255)
      else love.graphics.setColor(255, 215, 0, 255) end
      love.graphics.printf("Play to Combat", 555, 105, 140, "center")
    elseif moves[2]==true then  -- Play action
      drawgenericbutton(550, 90, 150, 84)
      if x>550 and x<700 and y>90 and y<174 then love.graphics.setColor(255, 255, 255, 255)
      else love.graphics.setColor(255, 215, 0, 255) end
      love.graphics.printf("Play Action", 555, 120, 140, "center")
    elseif moves[3]==true then  -- Attack
      drawgenericbutton(550, 90, 150, 84)
      if x>550 and x<700 and y>90 and y<174 then love.graphics.setColor(255, 255, 255, 255)
      else love.graphics.setColor(255, 215, 0, 255) end
      love.graphics.printf("Attack", 555, 120, 140, "center")
    else end
    if moves[4]==true then      -- Use special ability
      drawgenericbutton(550, 174, 150, 84)
      if x>550 and x<700 and y>174 and y<258 then love.graphics.setColor(255, 255, 255, 255)
      else love.graphics.setColor(50, 0, 200, 255) end
      love.graphics.printf("Use Special Ability", 555, 189, 140, "center")
    end
    if moves[5]==true then      -- Target
      drawgenericbutton(550, 258, 150, 84)
      if x>550 and x<700 and y>258 and y<342 then love.graphics.setColor(255, 255, 255, 255)
      else love.graphics.setColor(255, 215, 0, 255) end
      love.graphics.printf("Target", 555, 288, 140, "center")
    end
    if moves[6]==true then      -- Sacrifice
      drawgenericbutton(550, 342, 150, 84)
      if x>550 and x<700 and y>342 and y<426 then love.graphics.setColor(255, 255, 255, 255)
      else love.graphics.setColor(140, 0, 0, 255) end
      love.graphics.printf("Sacrifice Card", 555, 357, 140, "center")
    end
  end
end

function drawheldcard()
  local x, y = love.mouse.getX() / gRATIOX, love.mouse.getY() / gRATIOY
  if gDRAGGING == true and gHOLDING > (gFPS*0.3) and next(tCLICK)~=nil then drawcard(tCLICK[1].Card, x, y, 50, 70, 0, 100) end
end

function drawfilemenu()
  local x, y = love.mouse.getX() / gRATIOX, love.mouse.getY() / gRATIOY
  local i = 0
  local n
  if next(tSAVES)~=nil then n=#tSAVES else n=0 end
  if gFILEMODE == "Load" then drawtitle("Load Deck")
  elseif gFILEMODE == "Save" then drawtitle("Save Deck")
  elseif gFILEMODE == "Erase" then drawtitle("Erase Deck")
  else end
  drawcointotal()
  drawdeckmenubutton()
  drawminideck(gTURN)
  if gFILEMODE=="Save" then drawgenericbutton(150, 60, 500, (math.min(n,10)*50)+50)
  else drawgenericbutton(150, 60, 500, math.min(n,10)*50) end
  love.graphics.setFont(fHEADING)
  if next(tSAVES)~=nil then
    repeat
      i = i + 1
      if x>160 and x<390 and y>((i-1)*50)+70 and y<(i*50)+50 and gTYPING==false then love.graphics.setColor(255, 255, 255, gFLASH)
      else love.graphics.setColor(255, 215, 0, 255) end
      love.graphics.printf(tSAVES[i], 160, ((i-1)*50)+70, 230, "left")
    until i==#tSAVES or i==10
    if i<#tSAVES then
      repeat
        i = i + 1
        if x>410 and x<640 and y>((i-11)*50)+70 and y<((i-10)*50)+50 and gTYPING==false then love.graphics.setColor(255, 255, 255, gFLASH)
        else love.graphics.setColor(255, 215, 0, 255) end
        love.graphics.printf(tSAVES[i], 410, ((i-11)*50)+70, 230, "left")
      until i==#tSAVES or i==19
    end
  end
  if gFILEMODE == "Save" then
    i = i + 1
    if i < 11 then
      if x>160 and x<390 and y>((i-1)*50)+70 and y<(i*50)+50 and gTYPING==false then love.graphics.setColor(255, 255, 255, gFLASH)
      else love.graphics.setColor(0, 255, 0, 255) end
      love.graphics.printf("Create new file...", 160, ((i-1)*50)+70, 230, "left")
    else
      if x>410 and x<640 and y>((i-11)*50)+70 and y<((i-10)*50)+50 and gTYPING==false then love.graphics.setColor(255, 255, 255, gFLASH)
      else love.graphics.setColor(0, 255, 0, 255) end
      love.graphics.printf("Create new file...", 410, ((i-11)*50)+70, 230, "left")
    end
  end
  if gTYPING == true then
    love.graphics.setColor(0, 0, 0, 200)
    love.graphics.rectangle("fill", 0, 0, 800, 600)
  end
end

function drawgenericbutton(x, y, width, height)
  love.graphics.setColor(0, 0, 0, 255)
  love.graphics.rectangle("fill", x, y, width, height)
  love.graphics.setColor(255, 255, 255, 255)
  love.graphics.rectangle("line", x, y, width, height)
end

function drawending()
  local i, j
  love.graphics.setFont(fTITLE)
  love.graphics.setColor(255, 215, 0, 255)
  if gethealth(tHERO[1])>0 and gethealth(tHERO[2])<=0 then love.graphics.printf("P1 wins", 0, 40, 800, "center")
  elseif gethealth(tHERO[2])>0 and gethealth(tHERO[1])<=0 then love.graphics.printf("P2 wins", 0, 40, 800, "center")
  else love.graphics.printf("It's a draw", 0, 40, 800, "center") end
  love.graphics.printf("P1", 0, 150, 200, "center")
  love.graphics.printf("P2", 600, 150, 200, "center")
  if tCONTROL[2]=="AI" and gethealth(tHERO[2])<=0 and gLEVEL_CURSOR~=0 then
    love.graphics.setColor(0, 255, 0, 255)
    if gDIFFICULTY==1 then love.graphics.printf("Wins $"..tHABITAT[gLEVEL_CURSOR].Easy, 0, 200, 200, "center")
    elseif gDIFFICULTY==2 then love.graphics.printf("Wins $"..tHABITAT[gLEVEL_CURSOR].Medium, 0, 200, 200, "center")
    elseif gDIFFICULTY==3 then love.graphics.printf("Wins $"..tHABITAT[gLEVEL_CURSOR].Hard, 0, 200, 200, "center")
    else end
  end
  drawcard(tHERO[1], 300, 205, 150, 210)
  drawcard(tHERO[2], 500, 205, 150, 210)
  if next(tUNLOCKED)~=nil then
    love.graphics.setFont(fHEADING)
    love.graphics.setColor(255, 255, 255, 255)
    love.graphics.printf("You have unlocked these cards", 0, 325, 800, "center")
    for i = 1, #tUNLOCKED, 1 do
      j = tUNLOCKED[i]
      drawcard(tCARDS[j], i*200, 480, 150, 210)
    end
  end
end

function drawgameinfo()
  local x, y = love.mouse.getX() / gRATIOX, love.mouse.getY() / gRATIOY
  local yINFO = {315, 165}
  local i, j
  love.graphics.setFont(fHEADING)
  for i = 1, 2, 1 do
    if i==gTURN then love.graphics.setColor(255, 215, 0, gFLASH)
    else love.graphics.setColor(255, 215, 0, 255) end
    love.graphics.printf("P"..i, 10, yINFO[i], 120, "center")
    if next(tHERO[i])~=nil then
      if gethealth(tHERO[i])>tHERO[i].MaxHealth*0.7 then love.graphics.setColor(0, 255, 0, 255)
      elseif gethealth(tHERO[i])>tHERO[i].MaxHealth*0.4 then love.graphics.setColor(255, 128, 0, 255)
      else love.graphics.setColor(255, 0, 0, 255) end
      love.graphics.printf("HP: "..gethealth(tHERO[i]).."/"..tHERO[i].MaxHealth, 10, yINFO[i]+30, 125, "left")
    end
    if next(tBUDGET)~=nil and next(tPOINTS)~=nil then
      if tPOINTS[i]>0 then love.graphics.setColor(255, 215, 0, 255)
      else love.graphics.setColor(50, 50, 50, 255) end
      love.graphics.printf("SP: "..tPOINTS[i].."/"..tBUDGET[i], 10, yINFO[i]+60, 125, "left")
    end
    if i==gTURN then
      if gDAYNIGHT>=6 then 
        love.graphics.setColor(255, 215, 0, 255)
        love.graphics.printf("Day "..(gDAYNIGHT-5).."/5", 10, yINFO[i]+90, 120, "left")
      else
        love.graphics.setColor(50, 0, 200, 255)
        love.graphics.printf("Night "..gDAYNIGHT.."/5", 10, yINFO[i]+90, 120, "left")
      end
    end 
  end
  if tCONTROL[gTURN]=="Player" and gZOOMED == false then
    love.graphics.setFont(fDEFAULT)
    i = xDECK[gTURN]+(gCARDWIDTH/2)+5
    j = yDECK[gTURN]-(gCARDHEIGHT/2)
    if gACTION=="" and next(tCLICK)==nil then  -- Draw end turn button
      drawgenericbutton(i, j, 70, 35)
      if x>i and x<i+70 and y>j and y<j+35 then love.graphics.setColor(0, 255, 0, 255)
      else love.graphics.setColor(80, 80, 80, 255) end
      love.graphics.printf("End Turn", i, j+10, 70, "center")
    end
    j = yDECK[gTURN]-18
    if gACTION~="" and next(tCLICK)~=nil then  -- Draw cancel button
      drawgenericbutton(i, j, 70, 35)   
      if x>i and x<i+70 and y>j and y<j+35 then love.graphics.setColor(255, 125, 0, 255)
      else love.graphics.setColor(80, 80, 80, 255) end
      love.graphics.printf("Cancel", i, j+10, 70, "center")
    end
    j = yDECK[gTURN]+(gCARDHEIGHT/2)-35
    if gACTION=="" and next(tCLICK)==nil then  -- Draw quit game button
      drawgenericbutton(i, j, 70, 35)
      if x>i and x<i+70 and y>j and y<j+35 then love.graphics.setColor(255, 0, 0, 255)
      else love.graphics.setColor(80, 80, 80, 255) end
      love.graphics.printf("Quit Game", i, j+10, 70, "center")
    end
  end
  if gACTION ~="" and AIsimulation==false then                        -- Print Action at top of screen         
    love.graphics.setColor(255, 255, 255, 255)
    love.graphics.printf(gACTION, 300, 0, 200, "center")
  end
end

function love.mousepressed(x, y, button, istouch)
  x = x / gRATIOX
  y = y / gRATIOY
  if gGAMEMODE == "Title" then
    gGAMEMODE = "Main Menu"
    gFADE = 255
  elseif gGAMEMODE == "Main Menu" then clickmainmenu(x,y)
  elseif gGAMEMODE == "Select Level" then clickselectlevel(x,y)
  elseif gGAMEMODE == "Deck Menu" then clickdeckmenu(x,y)
  elseif gGAMEMODE == "Deck Parade" then clickdeckparade(x,y)
  elseif gGAMEMODE == "File Menu" then clickfilemenu(x,y)
  elseif gGAMEMODE == "New Turn" then 
    gGAMEMODE = "Battle"
    gFADE = 255
  elseif gGAMEMODE == "Battle" and gZOOMED == false and gENDING == 0 then clickbattle(x,y)
  elseif gGAMEMODE == "Battle" and gZOOMED == true and gENDING == 0 then clickselectedcard(x,y)  
  else end
end

function love.mousereleased(x, y, button, istouch)
  x = x / gRATIOX
  y = y / gRATIOY
  if (gGAMEMODE == "Deck Parade" or gGAMEMODE == "Deck Menu") and gHOLDING > (gFPS*0.5) then releasedeck(x,y)
  elseif gGAMEMODE == "Battle" and gHOLDING > (gFPS*0.5) and gENDING == 0 then releasebattle(x,y)
  else end
  gHOLDING = 0
  gDRAGGING = false
end

function clickcard(card, source, player, index)
  local click = {}
  click.Card = card
  click.Source = source
  click.Player = player
  click.Index = index
  table.insert(tCLICK, 1, deepcopy(click))
  gDRAGGING = true
  gZOOMED = true
end

function clickmainmenu(x, y)
  local i
  if ((x>350 and x<520 and y>200 and y<225) or (x>280 and x<320 and y>190 and y<230)) then
    if getdeckcount(1)<gDECKSIZE then addpopup("You must first select a complete deck of cards")
    elseif next(tHERO[1])==nil then addpopup("You must select a hero for your deck")
    else
      gFADE = 255
      tCONTROL = {"Player", "AI"}
      if next(tHABITAT) ~= nil then
        for i = 1, #tHABITAT, 1 do
          if tHABITAT[i].Unlock > 0 then gLEVEL_CURSOR = i end
        end
        gGAMEMODE = "Select Level"
        gTURN = 1
      end
    end
  elseif (x>=350 and x<=420 and y>=275 and y<=300) or (x>=280 and x<=320 and y>=265 and y<=305) then
    gFADE = 255
    gTURN = 1
    gGAMEMODE = "Deck Menu"
    tCONTROL = {"Player", "AI"}
  elseif (x>=350 and x<=530 and y>=350 and y<=375) or (x>=280 and x<=320 and y>=340 and y<=380) then 
    gFADE = 255
    tCONTROL = {"Player", "Player"}
    gTURN = 1
    gGAMEMODE = "Deck Menu"
    addpopup("Player 1 select a deck")
  elseif (x>=350 and x<=400 and y>=425 and y<=450) or (x>=280 and x<=320 and y>=415 and y<=455) then 
    love.timer.sleep(0.5)
    love.event.quit()
  elseif x>=55 and x<=95 and y>=530 and y<=570 then
    if gSOUNDON == true then gSOUNDON = false else gSOUNDON = true end
  elseif x>=705 and x<=745 and y>=530 and y<=570 then
    if gMUSICON == true then gMUSICON = false else gMUSICON = true end
  else end
end

function clickselectlevel(x, y)
  local max = 0
  local width = 150
  local start, finish
  local clicked
  local i, j
  if next(tHABITAT) ~= nil then max = #tHABITAT end
  if x>=700 and y>=250 and y<=350 and gLEVEL_CURSOR<max then gLEVEL_CURSOR = gLEVEL_CURSOR + 1
  elseif x<=100 and y>=250 and y<=350 and gLEVEL_CURSOR>1 then gLEVEL_CURSOR = gLEVEL_CURSOR - 1
  elseif x>=0 and x<=180 and y>=0 and y<=50 then 
    gGAMEMODE = "Main Menu"
    gFADE = 255
  elseif next(tHABITAT) ~= nil then
    start = math.max(gLEVEL_CURSOR-2, 1)
    finish = math.min(gLEVEL_CURSOR+2, #tHABITAT)
    for i = start, finish, 1 do
      clicked=false
      j = i-gLEVEL_CURSOR
      if x>=325+(j*(width+10)) and x<=475+(j*(width+10)) and y>=415 and y<=445 and tHABITAT[i].Unlock>=1 then
        gDIFFICULTY = 1
        AImax_loops = 20
        clicked = true
      elseif x>=325+(j*(width+10)) and x<=475+(j*(width+10)) and y>=445 and y<=475 and tHABITAT[i].Unlock>=2 then
        gDIFFICULTY = 2
        AImax_loops = 80
        clicked = true
      elseif x>=325+(j*(width+10)) and x<=475+(j*(width+10)) and y>=475 and y<=505 and tHABITAT[i].Unlock>=3 then
        gDIFFICULTY = 3
        AImax_loops = 320
        clicked = true
      else end
      if clicked==true then
        gLEVEL_CURSOR = i
        autodeck(2, tHABITAT[i].Name)
        savedeck("Autosave", 1)
        refreshsavefiles()
        newgame()
      end
    end
  else end
end

function clickdeckmenu(x, y)
  local i, j, a, b, c
  if x>0 and x<180 and y>0 and y<50 then
    gGAMEMODE = "Main Menu"
    gFADE = 255
  elseif gZOOMED==true then clickcloseup(x,y)
  elseif x>700 and x<790 and y>70 and y<120 and getdeckcount(gTURN)>0 then  -- blank the deck
    tDECK[gTURN]={}
    tHERO[gTURN]={}
    addpopup("Your deck of cards has been emptied.")
  elseif x>700 and x<790 and y>135 and y<185 and next(tSAVES)~=nil then
    gGAMEMODE = "File Menu"
    gFILEMODE = "Load"
    gFADE = 255
  elseif x>700 and x<790 and y>200 and y<250 and getdeckcount(gTURN)==gDECKSIZE and next(tHERO[gTURN])~=nil then
    gGAMEMODE = "File Menu"
    gFILEMODE = "Save"
    gFADE = 255
  elseif x>700 and x<790 and y>265 and y<315 then
    autodeck(gTURN, "All")
    sortdeck(gTURN)
  elseif x>700 and x<790 and y>330 and y<380 and next(tSAVES)~=nil then
    gGAMEMODE = "File Menu"
    gFILEMODE = "Erase"
    gFADE = 255
  elseif x>700 and x<790 and y>480 and y<530 and next(tHERO[gTURN])~=nil and getdeckcount(gTURN)==gDECKSIZE then
    if gTURN==1 then
      if tCONTROL[2]=="Player" then 
        gTURN = 2
        savedeck("Autosave", 1)
        refreshsavefiles()
      else
        gGAMEMODE = "Main Menu"
        gFADE = 255
      end
    else
      newgame()
    end
  elseif x>50 and x<700 and y>60 and y<480 then
    a, b, c, i = 10, 60, 0, 0
    if next(tHABITAT) ~= nil then
      repeat
        i = i + 1
        if x>a and x<a+120 and y>b and y<b+120 and tHABITAT[i].Unlock > 0 then c = i end
        a = a + 140
        if a > 630 then
          a = 10
          b = b + 145
        end
      until i==#tHABITAT or i==15
      if c ~= 0 then
        gFADE = 255
        gGAMEMODE = "Deck Parade"
        tLIST = deepcopy(getlistbyhabitat(tHABITAT[c].Name))
        gDECK_SELECT = c
        tCLICK = {}
        gZOOMED = false
      end
      j = #tHABITAT
    end
    j = i
    i, c = 0, 0
    if next(tSAVES)~=nil and j<15 then
      repeat
        i = i + 1
        if x>a and x<a+120 and y>b and y<b+120 then c = i end
        a = a + 140
        if a >630 then
          a = 10
          b = b + 145
        end
      until i==#tSAVES or j+i==15
      if c ~= 0 then
        loaddeck(tSAVES[c], gTURN)
      end
    end
  else end
  clickminideck(x,y)
end

function clickcloseup(x,y)
  card = getcardbyid(tCLICK[1].Card.ID)
  if x>550 and x<700 and y>60 and y<140 and gCOINS>=tCLICK[1].Card.Price and tCLICK[1].Card.Copies<tCLICK[1].Card.Maximum and tCLICK[1].Source=="List" then
    tCARDS[card].Copies = tCARDS[card].Copies + 1
    gCOINS = gCOINS - tCLICK[1].Card.Price
    savecards()
    --play sound ka-ching
    tLIST = deepcopy(getlistbyhabitat(tCLICK[1].Card.Habitat))
  elseif x>550 and x<700 and y>260 and y<340 and getcountindeckbyid(tCLICK[1].Card.ID, gTURN)<tCLICK[1].Card.Copies and tCLICK[1].Source=="List" then
    addcardtodeck(tCLICK[1].Card, gTURN)
    sortdeck(gTURN)
  elseif x>100 and x<250 and y>260 and y<340 and getcountindeckbyid(tCLICK[1].Card.ID, gTURN)>0 then
    removecardfromdeck(tCLICK[1].Card, gTURN)
    if tCLICK[1].Source=="Deck" then
      tCLICK={}
      gZOOMED=false
    end
  elseif x>250 and x<550 and y>60 and y<480 then -- do nothing, clicked zoomed in card
  else
    tCLICK={}
    gZOOMED=false
  end
end

function clickdeckparade(x,y)
  local a, b, c, i
  if x>=0 and x<=180 and y>=0 and y<=50 then
    gGAMEMODE = "Deck Menu"
    gFADE = 255
    tCLICK = {}
    gZOOMED = false
  elseif gZOOMED==true then clickcloseup(x,y)
  elseif next(tLIST)~=0 then
    a, b, c, i = 0, 60, 0, 0
    repeat
      i = i + 1
      if x>a and x<a+110 and y>b and y<b+154 then c = i end
      a = a + 115
      if a > 690 then
        a = 0
        b = b + 160
      end
    until i==#tLIST or i==21
    if c~=0 then clickcard(tLIST[c], "List", gTURN, c) end
  else end
  clickminideck(x,y)  
end

function clickfilemenu(x,y)
  local f = 0
  local n
  if next(tSAVES)~=nil then n = #tSAVES else n = 0 end
  if x>=0 and x<=180 and y>=0 and y<=50 then
    gGAMEMODE = "Deck Menu"
    gFADE = 255
  elseif x>160 and x<390 and y>70 and y<550 and gTYPING==false then
    f = math.floor((y-20)/50)
  elseif x>410 and x<540 and y>70 and y<550 and gTYPING==false then
    f = math.floor((y-20)/50)+10
  elseif x>300 and x<500 and y>280 and y<320 and gTYPING==true and gTYPEDTEXT=="Enter name of deck..." then
    gTYPEDTEXT = ""
  else end
  if f>0 and gTYPING==false then
    if f<=n then
      if gFILEMODE=="Load" then
        loaddeck(tSAVES[f],gTURN)
      elseif gFILEMODE=="Save" then
        savedeck(tSAVES[f],gTURN)
      elseif gFILEMODE=="Erase" then
        erasedeck(tSAVES[f])
        refreshsavefiles()
      else end
    elseif f==n+1 and gFILEMODE=="Save" then
      gTYPING=true
      gTYPEDTEXT="Enter name of deck..."
    else end
  end
end

function clickminideck(x,y)
  local i = 0
  if y>560 and y<600 then
    if x>0 and x<40 and next(tHERO[gTURN])~=nil then
      clickcard(tHERO[gTURN], "Hero", gTURN, gTURN)
    elseif x>40 and x<800 and next(tDECK[gTURN])~=nil then
      i = math.min(math.ceil((x-40)/19),39)
      if i<=#tDECK[gTURN] then 
        clickcard(tDECK[gTURN][i], "Deck", gTURN, i)
      end
    else end
  end
end

function clickbattle(x,y)
  local i, j, moves, targets
  local clear = false
  for i = 1, 2, 1 do
    if next(tDECK[i])~=nil then
      if tDECK[i][1].Hover==true then 
        tDECK[i][1].Hover=false
        clickcard(tDECK[i][1], "Deck", i, 1)
      end
    end
    if next(tHAND[i])~=nil then
      for j = #tHAND[i], 1, -1 do
        if tHAND[i][j].Hover==true then
          tHAND[i][j].Hover=false
          clickcard(tHAND[i][j], "Hand", i, j)
          moves = calculatemoves(gTURN)
          if love.keyboard.isDown("p") and moves[1]==true then        -- Shortcut to play to combat (click-p)
            tPOINTS[gTURN] = tPOINTS[gTURN] - tCLICK[1].Card.Tier
            addpopup("P"..gTURN.." plays "..tHAND[i][j].Name.." to combat")
            sendcombat(tHAND[i][j], tCLICK[1].Source, tCLICK[1].Player, tCLICK[1].Index)
            tCLICK = {}
            gZOOMED = false
          elseif love.keyboard.isDown("p") and moves[2]==true then    -- Shortcut to play action (click-p)
            gACTION = tHAND[i][j].Special
            if playaction(tHAND[i][j], gTURN) == true then
              tPOINTS[gTURN] = tPOINTS[gTURN] - tCLICK[1].Card.Tier
              targets = tHAND[i][j].SpTargets
              sendgraveyard(tHAND[i][j], tCLICK[1].Source, tCLICK[1].Player, tCLICK[1].Index)
              if targets == 0 and next(tCLICK)~=nil then
                table.remove(tCLICK,1)
                gACTION = ""
              end
            else 
              gACTION = ""
              tCLICK = {}
            end
            gZOOMED = false
          elseif love.keyboard.isDown("s") and moves[6]==true then    -- Shortcut to sacrifice card (click-s)
            addpopup("P"..gTURN.." sacrifices "..tHAND[i][j].Name)
            sendgraveyard(tHAND[i][j], tCLICK[1].Source, tCLICK[1].Player, tCLICK[1].Index)
            tBUDGET[gTURN] = tBUDGET[gTURN] + 1
            tPOINTS[gTURN] = tPOINTS[gTURN] + 1 
            gSACRIFICED = true
            tCLICK = {}
            gZOOMED = false
          else end
        end 
      end
    end
    if next(tGRAVE[i])~=nil then
      if tGRAVE[i][#tGRAVE[i]].Hover==true then
        tGRAVE[i][#tGRAVE[i]].Hover=false
        clickcard(tGRAVE[i][#tGRAVE[i]], "Grave", i, #tGRAVE[i])
      end
    end
    if next(tCOMBAT[i])~=nil then
      for j = #tCOMBAT[i], 1, -1 do
        if tCOMBAT[i][j].Hover==true then
          tCOMBAT[i][j].Hover=false
          clickcard(tCOMBAT[i][j], "Combat", i, j)
          moves = calculatemoves(gTURN)
          if love.keyboard.isDown("a") and moves[3]==true then        -- Shortcut to attack (click-a)
            gACTION = "Attack"
            if findtargets() then
              addpopup("Choose target")
            else
              gACTION = ""
              tCLICK = {}
              addpopup("No valid targets")
            end
            gZOOMED = false
          elseif love.keyboard.isDown("u") and moves[4]==true then    -- Shortcut to special ability (click-u)
            gACTION = tCOMBAT[i][j].Special
            if playaction(tCOMBAT[i][j], gTURN) == true then
              tPOINTS[gTURN] = tPOINTS[gTURN] - tCLICK[1].Card.SpVal
              tCOMBAT[i][j].SpecialUsed = true
              if tCOMBAT[i][j].SpTargets == 0 and next(tCLICK)~=nil then
                table.remove(tCLICK,1)
                gACTION = ""
              end
            else 
              gACTION = ""
              tCLICK = {}
            end
            gZOOMED = false
          elseif love.keyboard.isDown("t") and moves[5]==true then    -- Shortcut to target card (click-t)
            if gACTION == "Attack" or #tCLICK-1 >= tCLICK[#tCLICK].Card.SpTargets then clear = true end
            dotarget()
            if clear == true then 
              gACTION= ""
              tCLICK = {}
              cleartargets()
            end
            gZOOMED = false
          else end
        end
      end
    end
    if next(tHERO[i])~=nil then
      if tHERO[i].Hover==true then
        tHERO[i].Hover=false
        clickcard(tHERO[i], "Hero", i, i)
        moves = calculatemoves(gTURN)
        if love.keyboard.isDown("a") and moves[3]==true then        -- Shortcut to attack (click-a)
          gACTION = "Attack"
          if findtargets() then
            addpopup("Choose target")
          else
            gACTION = ""
            tCLICK = {}
            addpopup("No valid targets")
          end
          gZOOMED = false
        elseif love.keyboard.isDown("u") and moves[4]==true then    -- Shortcut to special ability (click-u)
          gACTION = tHERO[i].Special
          if playaction(tHERO[i], gTURN) == true then
            tPOINTS[gTURN] = tPOINTS[gTURN] - tCLICK[1].Card.SpVal
            tHERO[i].SpecialUsed = true
            if tHERO[i].SpTargets == 0 and next(tCLICK)~=nil then
              table.remove(tCLICK,1)
              gACTION = ""
            end
          else 
            gACTION = ""
            tCLICK = {}
          end
          gZOOMED = false
        elseif love.keyboard.isDown("t") and moves[5]==true then    -- Shortcut to target card (click-t)
          if gACTION == "Attack" or #tCLICK-1 == tCLICK[#tCLICK].Card.SpTargets then
            clear = true
          end
          dotarget()
          if clear == true then 
            gACTION = ""
            tCLICK = {}
            cleartargets()
          end
          gZOOMED = false
        else end
      end
    end
  end
  i = xDECK[gTURN]+(gCARDWIDTH/2)+5
  j = yDECK[gTURN]-(gCARDHEIGHT/2)
  if gACTION=="" and next(tCLICK)==nil and x>i and x<i+70 and y>j and y<j+35 then     -- Clicked end turn
    gTURN = 3 - gTURN
    newturn(gTURN)
  end
  j = yDECK[gTURN]-18
  if gACTION~="" and next(tCLICK)~=nil and x>i and x<i+70 and y>j and y<j+35 then     -- Clicked cancel move
    gACTION = ""
    tCLICK = {}
  end
  j = yDECK[gTURN]+(gCARDHEIGHT/2)-35
  if gACTION=="" and next(tCLICK)==nil and x>i and x<i+70 and y>j and y<j+35 then     -- Clicked quit game
    resetmainmenu()
    gGAMEMODE = "Main Menu"
  end
end

function clickselectedcard(x,y)
  local clear = false
  local card, moves, targets
  if next(tCLICK)~=nil then
    if tCLICK[1].Source == "Hand" then card = tHAND[tCLICK[1].Player][tCLICK[1].Index]
    elseif tCLICK[1].Source == "Deck" then card = tDECK[tCLICK[1].Player][tCLICK[1].Index]
    elseif tCLICK[1].Source == "Grave" then card = tGRAVE[tCLICK[1].Player][tCLICK[1].Index]
    elseif tCLICK[1].Source == "Combat" then card = tCOMBAT[tCLICK[1].Player][tCLICK[1].Index]
    elseif tCLICK[1].Source == "Hero" then card = tHERO[tCLICK[1].Player]
    else end
  end
  if gZOOMED == true and gDRAGGING == false then
    moves = calculatemoves(gTURN)
    if moves[1]==true and x>550 and x<700 and y>90 and y<174 then       -- March ally
      tPOINTS[gTURN] = tPOINTS[gTURN] - tCLICK[1].Card.Tier
      addpopup("P"..gTURN.." plays "..card.Name.." to combat")
      sendcombat(card, tCLICK[1].Source, tCLICK[1].Player, tCLICK[1].Index)
      tCLICK = {}
      gZOOMED = false
    elseif moves[2]==true and x>550 and x<700 and y>90 and y<174 then   -- Play action
      gACTION = card.Special
      if playaction(card, gTURN) == true then
        tPOINTS[gTURN] = tPOINTS[gTURN] - tCLICK[1].Card.Tier
        targets = card.SpTargets
        sendgraveyard(card, tCLICK[1].Source, tCLICK[1].Player, tCLICK[1].Index)
        if targets == 0 and next(tCLICK)~=nil then
          table.remove(tCLICK,1)
          gACTION = ""
        end
      else
        gACTION = ""
        tCLICK = {}
      end
      gZOOMED = false
    elseif moves[3]==true and x>550 and x<700 and y>90 and y<174 then   -- Attack
      gACTION = "Attack"
      if findtargets() then
        addpopup("Choose target")
      else
        gACTION = ""
        tCLICK = {}
        addpopup("No valid targets")
      end
      gZOOMED = false
    elseif moves[4]==true and x>550 and x<700 and y>174 and y<258 then  -- Use special ability
      gACTION = card.Special
      if playaction(card, gTURN) == true then
        tPOINTS[gTURN] = tPOINTS[gTURN] - tCLICK[1].Card.SpVal
        card.SpecialUsed = true
        if card.SpTargets == 0 and next(tCLICK)~=nil then
          table.remove(tCLICK,1)
          gACTION = ""
        end
      else 
        gACTION = ""
        tCLICK = {}
      end
      gZOOMED = false
    elseif moves[5]==true and x>550 and x<700 and y>258 and y<342 then  -- Target this card
      if gACTION == "Attack" or #tCLICK-1 == tCLICK[#tCLICK].Card.SpTargets then clear = true end
      dotarget()
      if clear == true then
        gACTION = ""
        tCLICK = {}
        cleartargets()
      end
      gZOOMED = false
    elseif moves[6]==true and x>550 and x<700 and y>342 and y<426 then  -- Sacrifice card
      addpopup("P"..gTURN.." sacrifices "..card.Name)
      sendgraveyard(card, tCLICK[1].Source, tCLICK[1].Player, tCLICK[1].Index)
      tBUDGET[gTURN] = tBUDGET[gTURN] + 1
      tPOINTS[gTURN] = tPOINTS[gTURN] + 1 
      gSACRIFICED = true
      tCLICK = {}
      gZOOMED = false
    elseif x>250 and x<550 and y>90 and y<510 then                      -- Do nothing, clicked the zoomed in card
      
    else                                                                -- Zoom out and remove this card from tCLICK
      gZOOMED = false
      if next(tCLICK)~=nil then table.remove(tCLICK,1) end
    end
  end
end

function releasedeck(x,y)
  if next(tCLICK)~=nil then
    if tCLICK[1].Source == "Deck" and y < 540 then
      removecardfromdeck(tCLICK[1].Card, gTURN)
      tCLICK = {}
      gZOOMED = false
    elseif tCLICK[1].Source == "List" and y > 540 and getcountindeckbyid(tCLICK[1].Card.ID, gTURN)<tCLICK[1].Card.Copies then
      addcardtodeck(tCLICK[1].Card, gTURN)
      sortdeck(gTURN)
      tCLICK = {}
      gZOOMED = false
    else end
  end
end

function releasebattle(x,y)
  local a = gTURN
  local b = 3 - gTURN
  local card, moves, targets
  local i
  if next(tCLICK)~=nil then
    if tCLICK[1].Source == "Hand" then card = tHAND[tCLICK[1].Player][tCLICK[1].Index]
    elseif tCLICK[1].Source == "Deck" then card = tDECK[tCLICK[1].Player][tCLICK[1].Index]
    elseif tCLICK[1].Source == "Grave" then card = tGRAVE[tCLICK[1].Player][tCLICK[1].Index]
    elseif tCLICK[1].Source == "Combat" then card = tCOMBAT[tCLICK[1].Player][tCLICK[1].Index]
    elseif tCLICK[1].Source == "Hero" then card = tHERO[tCLICK[1].Player]
    else end
  end
  if gDRAGGING == true and next(tCLICK)~=nil then
    moves = calculatemoves(a)
    if moves[1]==true and x>xCOMBAT[a]-(gCOMBATWIDTH/2) and x<xCOMBAT[a]+(gCOMBATWIDTH/2) and y>yCOMBAT[a]-(gCARDHEIGHT/2) and y<yCOMBAT[a]+(gCARDHEIGHT/2) then   -- March ally
      tPOINTS[a] = tPOINTS[a] - tCLICK[1].Card.Tier
      addpopup("P"..gTURN.." plays "..card.Name.." to combat")
      sendcombat(card, tCLICK[1].Source, tCLICK[1].Player, tCLICK[1].Index)
      tCLICK = {}
      gZOOMED = false
    elseif moves[2]==true and x>xCOMBAT[a]-(gCOMBATWIDTH/2) and x<xCOMBAT[a]+(gCOMBATWIDTH/2) and y>yCOMBAT[a]-(gCARDHEIGHT/2) and y<yCOMBAT[a]+(gCARDHEIGHT/2) then   -- Play action
      gACTION = card.Special
      if playaction(card, a) == true then
        tPOINTS[a] = tPOINTS[a] - tCLICK[1].Card.Tier
        targets = card.SpTargets
        sendgraveyard(card, tCLICK[1].Source, tCLICK[1].Player, tCLICK[1].Index)
        if targets == 0 and next(tCLICK)~=nil then
          table.remove(tCLICK,1)
          gACTION = ""
        end
        gZOOMED = false
      else gACTION = "" end
    elseif moves[3]==true and x>xCOMBAT[b]-(gCOMBATWIDTH/2) and x<xCOMBAT[b]+(gCOMBATWIDTH/2) and y>yCOMBAT[b]-(gCARDHEIGHT/2) and y<yCOMBAT[b]+(gCARDHEIGHT/2) then   -- Are they attacking an ally?
      gACTION = "Attack"
      if findtargets() then
        for i = #tCOMBAT[b], 1, -1 do
          if tCOMBAT[b][i].Hover == true and tCOMBAT[b][i].Target == true then
            clickcard(tCOMBAT[b][i], "Combat", b, i)
            dotarget()
          end
        end
      end
      tCLICK = {}
      gACTION = ""
      gZOOMED = false
     elseif moves[3]==true and x>xHERO[b]-(gCARDWIDTH/2) and x<xHERO[b]+(gCARDWIDTH/2) and y>yHERO[b]-(gCARDHEIGHT/2) and y<yHERO[b]+(gCARDHEIGHT/2) then   -- Are they attacking the hero?
      gACTION = "Attack"
      if findtargets() then
        if tHERO[b].Hover == true and tHERO[b].Target == true then
          clickcard(tHERO[b], "Hero", b, b)
          dotarget()
        end
      end
      tCLICK = {}
      gACTION = ""
      gZOOMED = false 
    elseif moves[6]==true and x>xGRAVE[gTURN]-(gCARDWIDTH/2) and x<xGRAVE[gTURN]+(gCARDWIDTH/2) and y>yGRAVE[gTURN]-(gCARDHEIGHT/2) and y<yGRAVE[gTURN]+(gCARDHEIGHT/2) then   -- Sacrifice card
      addpopup("P"..gTURN.." sacrifices "..card.Name)
      sendgraveyard(card, tCLICK[1].Source, tCLICK[1].Player, tCLICK[1].Index)
      tBUDGET[gTURN] = tBUDGET[gTURN] + 1
      tPOINTS[gTURN] = tPOINTS[gTURN] + 1 
      gSACRIFICED = true
      tCLICK = {}
      gZOOMED = false
    else end
  end
end

function love.keypressed(key)
  if gGAMEMODE == "Title" then
    gGAMEMODE = "Main Menu"
    gFADE = 255
  elseif gGAMEMODE == "Main Menu" then
    if key == "escape" then love.event.quit() end
  elseif gGAMEMODE == "Select Level" then
    if key == "escape" then 
      gFADE = 255
      gGAMEMODE = "Main Menu"
    end
  elseif gGAMEMODE == "Deck Menu" then
    if key == "escape" then
      gFADE = 255
      gGAMEMODE = "Main Menu"
    end
  elseif gGAMEMODE == "Deck Parade" then
    if key == "escape" then
      if gZOOMED==true then 
        tCLICK={}
        gZOOMED=false
      else
        gFADE = 255
        gGAMEMODE = "Deck Menu"
        tCLICK = {}
        gZOOMED = false
      end
    end
  elseif gGAMEMODE == "File Menu" then
    if key == "escape" then
      gFADE = 255
      gTYPING = false
      gTYPEDTEXT = ""
      gGAMEMODE = "Deck Menu"
    elseif key == "return" and gTYPING == true then
      if gTYPEDTEXT~="Enter name of deck..." and gTYPEDTEXT~="" then
        savedeck(gTYPEDTEXT, gTURN)
        refreshsavefiles()
        gTYPING=false
        gTYPEDTEXT=""
      end
    end
  elseif gGAMEMODE == "Battle" then
    if key == "escape" and tCONTROL[gTURN]=="Player" then
      gZOOMED = false
      tCLICK = {}
      gACTION = ""
    elseif key == "return" and tCONTROL[gTURN]=="Player" and gZOOMED == false and gACTION == "" then
      gTURN = 3 - gTURN
      newturn(gTURN)
    else end
  elseif gGAMEMODE == "New Turn" then
    gGAMEMODE = "Battle"
    gFADE = 255
  else end
end

function love.textinput(text)
  if gTYPING==true then
    if gTYPEDTEXT=="Enter name of deck..." then gTYPEDTEXT = "" end
    gTYPEDTEXT = gTYPEDTEXT..text
    if string.len(gTYPEDTEXT)>25 then gTYPEDTEXT=string.sub(gTYPEDTEXT,-25) end
  end
end

function getnumunlockedcards(habitat)
  local i
  local cards, unlocked = 0, 0
  if next(tCARDS) ~= nil then
    for i = 1, #tCARDS, 1 do
      if tCARDS[i].Habitat == habitat then
        cards = cards + 1
        if tCARDS[i].Unlocked == "Unlocked" then unlocked = unlocked + 1 end
      end
    end
  end
  return cards, unlocked
end

function getdeckcount(pl)
  if next(tDECK[pl]) ~= nil then return #tDECK[pl]
  else return 0 end
end

function gethandcount(pl)
  if next(tHAND[pl]) ~= nil then return #tHAND[pl]
  else return 0 end
end

function getcombatcount(pl)
  if next(tCOMBAT[pl]) ~= nil then return #tCOMBAT[pl]
  else return 0 end
end

function getcountindeckbyid(id, pl)
  local c = 0
  local i
  if next(tDECK[pl]) ~= nil then
    for i = 1, #tDECK[pl], 1 do
      if tDECK[pl][i].ID == id then c = c + 1 end
    end
  end
  if next(tHERO[pl]) ~= nil then
    if tHERO[pl].ID == id then c = c + 1 end
  end
  return c
end

function getcountindeckbytype(kind, pl)
  local c = 0
  local i
  if next(tDECK[pl]) ~= nil then
    for i = 1, #tDECK[pl], 1 do
      if tDECK[pl][i].Type == kind then c = c + 1 end
    end
  end
  if next(tHERO[pl]) ~= nil then
    if tHERO[pl].Type == kind then c = c + 1 end
  end
  return c
end

function getcountinhand(card, pl)
  local i, c = 0, 0
  if next(tHAND[pl])~=nil then
    for i = 1, #tHAND[pl], 1 do
      if tHAND[pl][i].ID == card.ID then c = c + 1 end
    end
  end
  return c
end

function gethealth(card)
  return card.MaxHealth - card.Wounds
end

function getarmour(card)
  local arm = 0
  if string.find(card.Special, "Armour +") then arm = arm + tonumber(string.sub(card.Special, 9)) end
  return arm
end

function getkilldmg(card)
  return card.MaxHealth + getarmour(card) - card.Wounds
end

function getpower(card)
  local p = card.Power
  if card.Effect[4]>0 then p = p * 0.5 end
  if card.Effect[5]>0 then p = p * 1.25 end
  if card.Effect[6]>0 then p = p * 1.33 end
  if card.Effect[7]>0 then p = p * 0.66 end
  if card.Effect[9]>0 then p = p * 1.5 end
  if card.Effect[10]>0 then p = p * 0.5 end
  if card.Effect[14]>0 then p = p * 1.25 end
  return math.floor(p+0.5)
end

function getprice(card)
  return math.floor(card.Price+0.5)
end

function addpopup(msg)
  if AIsimulation==false then
    if msg ~= "Choose target" and msg ~= "Choose targets" then print(msg) end
    gPOPUP_COUNT = gPOPUP_TIME
    gPOPUP_MSG = msg
  end
end

function addfloatytext(card, text)
  entry = {}
  entry.Text = text
  entry.X = card.X
  entry.Y = card.Y
  entry.Timer = 0
  table.insert(card.FloatyText, entry)
end

function endgame()
  local winner, loser = 0, 0
  local i, j
  if gethealth(tHERO[1])<=0 and gethealth(tHERO[2])>0 then 
    loser = 1
    winner = 2
  elseif gethealth(tHERO[2])<=0 and gethealth(tHERO[1])>0 then 
    loser = 2
    winner = 1
  else end
  if winner > 0 and loser > 0 then
    if tCONTROL[winner]=="Player" and tCONTROL[loser]=="AI" and gLEVEL_CURSOR ~= 0 then
      if gDIFFICULTY == 1 then 
        gCOINS = gCOINS + tHABITAT[gLEVEL_CURSOR].Easy
        if gLEVEL_CURSOR < #tHABITAT then tHABITAT[gLEVEL_CURSOR+1].Unlock = math.max(tHABITAT[gLEVEL_CURSOR+1].Unlock, 1) end
        unlockcard()
      elseif gDIFFICULTY == 2 then 
        gCOINS = gCOINS + tHABITAT[gLEVEL_CURSOR].Medium
        if gLEVEL_CURSOR < #tHABITAT then tHABITAT[gLEVEL_CURSOR+1].Unlock = math.max(tHABITAT[gLEVEL_CURSOR+1].Unlock, 2) end
        for i = 1, 2, 1 do
          unlockcard()
        end
      elseif gDIFFICULTY == 3 then 
        gCOINS = gCOINS + tHABITAT[gLEVEL_CURSOR].Hard
        if gLEVEL_CURSOR < #tHABITAT then tHABITAT[gLEVEL_CURSOR+1].Unlock = math.max(tHABITAT[gLEVEL_CURSOR+1].Unlock, 3) end
        for i = 1, 3, 1 do
          unlockcard()
        end
      else end
      savelevels()
      savecards()
    end
  end
end

function unlockcard()
  local locked = {}
  local i, j
  if next(tCARDS)~=nil and gLEVEL_CURSOR ~= 0 then
    for i = 1, #tCARDS, 1 do
      if tCARDS[i].Unlocked == "Locked" and tCARDS[i].Habitat == tHABITAT[gLEVEL_CURSOR].Name then table.insert(locked, i) end
    end
  end
  if next(locked)~=nil then
    i = math.random(#locked)
    j = locked[i]
    table.insert(tUNLOCKED, j)
    tCARDS[j].Unlocked = "Unlocked"
  end
end

function deepcopy(orig)
  local orig_type = type(orig)
  local copy
  if orig_type == 'table' then
    copy = {}
    for orig_key, orig_value in next, orig, nil do
      copy[deepcopy(orig_key)] = deepcopy(orig_value)
    end
    setmetatable(copy, deepcopy(getmetatable(orig)))
  else copy = orig end -- number, string, boolean, etc
  return copy
end

function AI()
  local backuptCLICK = deepcopy(tCLICK)
  local AIchoices = {}  -- the choices I have for moves
  local AIchain_start = {}  -- the move that started this chain
  local AIlayer = 0   -- counter for length of chain
  local report = ""
  local aiDECK, aiHAND, aiCOMBAT, aiGRAVE, aiHERO
  local aiBUDGET, aiPOINTS, aiSACRIFICED, aiCEASEFIRE, aiACTION
  local a = gTURN  -- friendly
  local b = 3 - gTURN  -- foe
  local i

  if AIloop < AImax_loops then  -- Still searching for moves?
    -- save the state of play
    aiDECK = deepcopy(tDECK)
    aiHAND = deepcopy(tHAND)
    aiCOMBAT = deepcopy(tCOMBAT)
    aiGRAVE = deepcopy(tGRAVE)
    aiHERO = deepcopy(tHERO)
    aiBUDGET = deepcopy(tBUDGET)
    aiPOINTS = deepcopy(tPOINTS)
    aiSACRIFICED = gSACRIFICED
    aiCEASEFIRE = gCEASEFIRE
    aiACTION = gACTION
    AIsimulation = true  -- just testing, not for real!
    gTIMER = love.timer.getTime()
    repeat  -- loop until this move chain comes to an end (or gets very long!)
      AIlayer = AIlayer + 1  -- counter for length of move chain
      AIchoices = listsomeAImoves()  -- list of moves I can make right now
      if next(AIchoices) ~= nil then  -- is there a move I can make?
        i = math.random(#AIchoices)  -- choose a move at random
        if AIlayer == 1 then AIchain_start = deepcopy(AIchoices[i]) end  -- remember the start
        report = AIreport(AIchoices[i])
        doAImove(AIchoices[i][1], AIchoices[i][2], AIchoices[i][3], AIchoices[i][4], AIchoices[i][5], AIchoices[i][6])  -- do the move
        if AIchoices[i][7] > AIbest_score then  -- check if it's a new best score
          print("#"..AIloop..": NEW BEST TARGET MOVE...")
          print("-"..report)
          AIbest_score = AIchoices[i][7]  -- new best score
          AIbest_start = deepcopy(AIchain_start)  -- recall how I got here first move..!
        end
      end
    until next(AIchoices) == nil or AIlayer > 10 or love.timer.getTime() - gTIMER > (2/gFPS)  -- until no choices / chain too long / takes too long!
    -- restore the state of play to what it was
    tDECK = deepcopy(aiDECK)
    tHAND = deepcopy(aiHAND)
    tCOMBAT = deepcopy(aiCOMBAT)
    tGRAVE = deepcopy(aiGRAVE)
    tHERO = deepcopy(aiHERO)
    tBUDGET = deepcopy(aiBUDGET)
    tPOINTS = deepcopy(aiPOINTS)
    gSACRIFICED = aiSACRIFICED
    gCEASEFIRE = aiCEASEFIRE
    gACTION = aiACTION
    AIsimulation = false  -- okay, we're not playing around anymore!
    AIloop = AIloop + 1  -- increase the counter towards making a move
  else  -- Time to move!
    if next(AIbest_start) ~= nil then  -- I've found something good to do, do the first bit!
      print(AIloop.." LOOPS COMPLETE. STARTING MOVE...")
      report = AIreport(AIbest_start)
      doAImove(AIbest_start[1], AIbest_start[2], AIbest_start[3], AIbest_start[4], AIbest_start[5], AIbest_start[6], AIbest_start[7])
      print("-"..report)
      print(tPOINTS[a].." points remaining.")
      AIbest_score = AIscore(a)  -- now current score is the one to beat
      AIbest_start = {}  -- and I have a blank slate to work with
      AIloop = 0  -- plenty of time to think again
      AIlastmove = love.timer.getTime()
    else  -- Nothing good found to do, end my turn
      gACTION = "End turn"
      print("No beneficial move left so end Player"..a.." turn")
      gTURN = b
      newturn(gTURN)
    end
  end
  tCLICK = deepcopy(backuptCLICK)
end

function AIscore(pl)
  local a = pl                  -- friendly
  local b = 3 - pl              -- foe
  local s = 0                   -- overall score
  local i, j                    -- loop variables, coeffs
  local c, p, h, x               -- general variables

  score_budget = 0              -- Used in AI() for feedback
  score_hand = 0                -- Used in AI() for feedback
  score_allies = 0              -- Used in AI() for feedback
  score_peace = 0               -- Used in AI() for feedback
  score_herohealth = 0          -- Used in AI() for feedback

  local AIbudget_ideal = 7      -- What's my ideal budget?
  local AIbudget_peak = 700     -- How much is that budget worth to me?
  local AIbudget_nominal = 5    -- How much is each budget point worth after that?
  local AIcard_minimum = 3      -- What's my lower card limit?
  local AIcard_peak = 900       -- How much is keeping at least that number of cards worth to me?
  local AIcard_nominal = 10     -- How much is each card worth after that?
  local AIcard_saver = 3        -- How far in advance will I wait for cards?
  local AIcard_penalty = 25     -- How much will I penalise overly expensive cards? (per point)
  local AIally_health = 0.9     -- How important is the health of allies?
  local AIally_power = 0.9      -- How important is the power of allies?
  local AIally_nominal = 100    -- How much is an ally's life worth to me?
  local AIally_defensive = 1    -- How much weight do I give to my defensive allies?
  local AIally_attacking = 1    -- How much weight do I give to their attacking allies?
  local AIself_preservation = 6 -- How much do I value keeping a single hero HP?
  local AIkiller_instinct = 7   -- How much do I value taking a single hero HP?

  -- BUDGETS
  i=AIbudget_peak/(AIbudget_ideal^2)
  j=i*AIbudget_ideal*2
  c=tBUDGET[a]
  if c <= AIbudget_ideal then score_budget=score_budget+((j*c)-(i*(c^2)))
  else score_budget=score_budget+AIbudget_peak end
  score_budget=score_budget+(c*AIbudget_nominal)
  c=tBUDGET[b]
  if c <= AIbudget_ideal then score_budget=score_budget-((j*c)-(i*(c^2)))
  else score_budget=score_budget-AIbudget_peak end
  score_budget=score_budget-(c*AIbudget_nominal)
  score_budget=math.floor(score_budget)
  -- NUMBER OF CARDS IN HAND
  i=AIcard_peak/(AIcard_minimum^2)
  j=i*AIcard_minimum*2
  c=#tHAND[a]
  if c <= AIcard_minimum then score_hand=score_hand+((j*c)-(i*(c^2)))
  else score_hand=score_hand+AIcard_peak end
  score_hand=score_hand+(c*AIcard_nominal)
  c=#tHAND[b]
  if c <= AIcard_minimum then score_hand=score_hand-((j*c)-(i*(c^2)))
  else score_hand=score_hand-AIcard_peak end
  score_hand=score_hand-(c*AIcard_nominal)
  -- USEFULNESS OF CARD IN HAND & MULTIPLES
  for i = 1, #tHAND[a], 1 do
    j=getcountinhand(tHAND[a][i], a)
    score_hand=score_hand-(25*(2^(j-1)))
    j=tHAND[a][i].Tier-tBUDGET[a]-AIcard_saver
    if j>0 then score_hand = score_hand-(j*AIcard_penalty) end
  end
  score_hand=math.floor(score_hand+500)  -- to counter negatives from duplicates
  -- DEFENSIVE ALLIES
  for i = 1, #tCOMBAT[a], 1 do
    score_allies = score_allies + AIally_nominal
    h = gethealth(tCOMBAT[a][i])
    p = getpower(tCOMBAT[a][i], a)
    if tCOMBAT[a][i].Effect[1]>0 then p=p*(0.75^(math.min(tCOMBAT[a][i].Effect[1],5))) end
    if tCOMBAT[a][i].Effect[2]>0 then p=p*1.1 end
    if tCOMBAT[a][i].Effect[3]>0 then p=p*(0.6^(math.min(tCOMBAT[a][i].Effect[3],5))) end
    x = 0
    if tCOMBAT[a][i].Effect[11]>0 then x = x + 1 end
    if tCOMBAT[a][i].Effect[12]>0 then x = x + 3 end
    if tCOMBAT[a][i].Effect[13]>0 then x = x + 5 end
    h = h * (math.max(h-x, 0) / h)
    score_allies = score_allies + (((h^AIally_health)*(p^AIally_power)) * AIally_defensive)
  end
  -- ATTACKING ALLIES
  for i = 1, #tCOMBAT[b], 1 do
    score_allies = score_allies - AIally_nominal
    h = gethealth(tCOMBAT[b][i])
    p = getpower(tCOMBAT[b][i], b)
    if tCOMBAT[b][i].Effect[1]>0 then p=p*(0.75^(math.min(tCOMBAT[b][i].Effect[1],5))) end
    if tCOMBAT[b][i].Effect[2]>0 then p=p*1.1 end
    if tCOMBAT[b][i].Effect[3]>0 then p=p*(0.6^(math.min(tCOMBAT[b][i].Effect[3],5))) end
    score_allies = score_allies - (((h^AIally_health)*(p^AIally_power)) * AIally_attacking)
  end
  if gCEASEFIRE > 0 then score_allies = score_allies / 2 end
  score_allies=math.floor(score_allies)
  -- HERO HP
  if next(tHERO[a]) ~= nil then
    score_herohealth = score_herohealth + AIself_preservation * gethealth(tHERO[a])
  end
  if next(tHERO[b]) ~= nil then
    score_herohealth = score_herohealth - AIkiller_instinct * gethealth(tHERO[b])
    if gethealth(tHERO[b]) <= 0 then score_herohealth = score_herohealth + 10000 end
  end
  score_herohealth=math.floor(score_herohealth)

  s = score_budget + score_hand + score_allies + score_peace + score_herohealth
  return s
end

function listsomeAImoves()
  local list = {}
  local move = {}
  local a = gTURN
  local b = 3 - gTURN
  local he = {0, 0}    -- hero
  local ha, co = 0, 0  -- hand_source, combat_source
  local ht = {0, 0}    -- hand_target
  local ct = {0, 0}    -- combat_target
  local special
  local i

  -- Pick a random card from everywhere available (otherwise zero)
  if next(tHAND[a]) ~= nil then ha = math.random(#tHAND[a]) end
  if next(tCOMBAT[a]) ~= nil then co = math.random(#tCOMBAT[a]) end
  for i = 1, 2, 1 do
    if next(tHERO[i]) ~= nil then he[i] = 1 end
    if next(tHAND[i]) ~= nil then ht[i] = math.random(#tHAND[i]) end
    if next(tCOMBAT[i]) ~= nil then ct[i] = math.random(#tCOMBAT[i]) end
  end

  checkpoint()     -- save the state of play

  if ha ~= 0 and gSACRIFICED ~= true then   -- Sacrifice an ally
    gACTION = "Sacrifice"  
    move={"Sacrifice", ha, "Hand", 0, "None", 0}
    if doAImove(move[1], move[2], move[3], move[4], move[5], move[6]) == true then
      table.insert(move, AIscore(a))
      table.insert(list, move)
    end  
    restorecheckpoint()
  end

  if ha ~= 0 then                           -- March an ally into battle
    if tHAND[a][ha].Type == "Ally" and tHAND[a][ha].Tier <= tPOINTS[a] and #tCOMBAT[a] < gMAXCARDSINCOMBAT then
      gACTION = "Play to Combat"  
      move={"Play to Combat", ha, "Hand", 0, "None", 0}
      if doAImove(move[1], move[2], move[3], move[4], move[5], move[6]) == true then
        table.insert(move, AIscore(a))
        table.insert(list, move)
      end
      restorecheckpoint()
    end
  end

  if co ~= 0 and gCEASEFIRE == 0 then       -- Attack using an ally
    if tCOMBAT[a][co].Used == false and tCOMBAT[a][co].Effect[1]==0 and tCOMBAT[a][co].Effect[3]==0 then
      gACTION = "Attack"  -- Attack a target (ally or hero)
      if findtargets() ~= false then
        harfway()
        if ct[b] ~= 0 then    -- Attack an ally
          if tCOMBAT[b][ct[b]].Target == true then
            move={"Attack", co, "Combat", ct[b], "Combat", b}
            if doAImove(move[1], move[2], move[3], move[4], move[5], move[6]) == true then
              table.insert(move, AIscore(a))
              table.insert(list, move)
            end
            restoreharfway()
          end
        end
        if he[b] ~= 0 then    -- Attack a hero
          if tHERO[b].Target == true then
            move={"Attack", co, "Combat", 1, "Hero", b}
            if doAImove(move[1], move[2], move[3], move[4], move[5], move[6]) == true then
              table.insert(move, AIscore(a))
              table.insert(list, move)
            end
            restoreharfway()
          end
        end
      end
      restorecheckpoint()
    end
  end

  if ha ~= 0 then   -- Play an action card from hand (on any target or none)
    if tHAND[a][ha].Type == "Action" and tHAND[a][ha].Tier <= tPOINTS[a] then
      gACTION = tHAND[a][ha].Special
      if tHAND[a][ha].SpTargets == 0 then
        move={"Play action", ha, "Hand", 0, "None", 0}
        if doAImove(move[1], move[2], move[3], move[4], move[5], move[6]) == true then
          table.insert(move, AIscore(a))
          table.insert(list, move)
        end
      else
        if findtargets() ~= false then
          harfway()
          for i = 1, 2, 1 do
            if ht[i] ~= 0 then
              if tHAND[i][ht[i]].Target == true then
                move={"Play action", ha, "Hand", ht[i], "Hand", i}
                if doAImove(move[1], move[2], move[3], move[4], move[5], move[6]) == true then
                  table.insert(move, AIscore(a))
                  table.insert(list, move)
                end
                restoreharfway()
              end
            end
            if ct[i] ~= 0 then
              if tCOMBAT[i][ct[i]].Target == true then
                move={"Play action", ha, "Hand", ct[i], "Combat", i}
                if doAImove(move[1], move[2], move[3], move[4], move[5], move[6]) == true then
                  table.insert(move, AIscore(a))
                  table.insert(list, move)
                end
                restoreharfway()
              end
            end
            if he[i] ~= 0 then
              if tHERO[i].Target == true then
                move={"Play action", ha, "Hand", 1, "Hero", i}
                if doAImove(move[1], move[2], move[3], move[4], move[5], move[6]) == true then
                  table.insert(move, AIscore(a))
                  table.insert(list, move)
                end
                restoreharfway()
              end
            end
          end
        end
      end
      restorecheckpoint()
    end
  end

  if co ~= 0 then   -- Use a special ability belonging to an ally (on any target or none)
    if tCOMBAT[a][co].Special ~= "None" and tCOMBAT[a][co].SpVal > 0 and tCOMBAT[a][co].SpVal <= tPOINTS[a] and tCOMBAT[a][co].SpecialUsed ~= true then
      gACTION = tCOMBAT[a][co].Special
      if tCOMBAT[a][co].SpTargets == 0 then
        move={"Use", co, "Combat", 0, "None", 0}
        if doAImove(move[1], move[2], move[3], move[4], move[5], move[6]) == true then
          table.insert(move, AIscore(a))
          table.insert(list, move)
        end
      else
        if findtargets() ~= false then
          harfway()
          for i = 1, 2, 1 do
            if ct[i] ~= 0 then
              if tCOMBAT[i][ct[i]].Target == true then
                move={"Use", co, "Combat", ct[i], "Combat", i}
                if doAImove(move[1], move[2], move[3], move[4], move[5], move[6]) == true then
                  table.insert(move, AIscore(a))
                  table.insert(list, move)
                end
                restoreharfway()
              end
            end
            if ht[i] ~= 0 then
              if tHAND[i][ht[i]].Target == true then
                move={"Use", co, "Combat", ht[i], "Hand", i}
                if doAImove(move[1], move[2], move[3], move[4], move[5], move[6]) == true then
                  table.insert(move, AIscore(a))
                  table.insert(list, move)
                end
                restoreharfway()
              end
            end
            if he[i] ~= nil then
              if tHERO[i].Target == true then
                move={"Use", co, "Combat", 1, "Hero", i}
                if doAImove(move[1], move[2], move[3], move[4], move[5], move[6]) == true then
                  table.insert(move, AIscore(a))
                  table.insert(list, move)
                end
                restoreharfway()
              end
            end
          end
        end
      end
      restorecheckpoint()
    end
  end

  if he[a] ~= 0 then   -- Use a special ability belonging to a hero (on any target or none)
    if tHERO[a].Special ~= "None" and tHERO[a].SpVal > 0 and tHERO[a].SpVal <= tPOINTS[a] and tHERO[a].SpecialUsed ~= true then
      gACTION = tHERO[a].Special
      if tHERO[a].SpTargets == 0 then
        move={"Use", 1, "Hero", 0, "None", 0}
        if doAImove(move[1], move[2], move[3], move[4], move[5], move[6]) == true then
          table.insert(move, AIscore(a))
          table.insert(list, move)
        end
      else
        if findtargets() ~= false then
          harfway()
          for i = 1, 2, 1 do
            if ct[i] ~= 0 then
              if tCOMBAT[i][ct[i]].Target == true then
                move={"Use", 1, "Hero", ct[i], "Combat", i}
                if doAImove(move[1], move[2], move[3], move[4], move[5], move[6]) == true then
                  table.insert(move, AIscore(a))
                  table.insert(list, move)
                end
                restoreharfway()
              end
            end
            if ht[i] ~= 0 then
              if tHAND[i][ht[i]].Target == true then
                move={"Use", 1, "Hero", ht[i], "Hand", i}
                if doAImove(move[1], move[2], move[3], move[4], move[5], move[6]) == true then
                  table.insert(move, AIscore(a))
                  table.insert(list, move)
                end
                restoreharfway()
              end
            end
            if he[i] ~= nil then
              if tHERO[i].Target == true then
                move={"Use", 1, "Hero", 1, "Hero", i}
                if doAImove(move[1], move[2], move[3], move[4], move[5], move[6]) == true then
                  table.insert(move, AIscore(a))
                  table.insert(list, move)
                end
                restoreharfway()
              end
            end
          end
        end
      end
      restorecheckpoint()
    end
  end

  return list
end

function doAImove(move, origin, origin_home, target, target_home, target_player)
  local a = gTURN
  local b = 3 - gTURN
  local r = false
  local origin_card = {}
  local target_card = {}
  local click = {}

  tCLICK = {}
  gACTION = move
  
  if origin_home == "Hand" and next(tHAND[a][origin]) ~= nil then origin_card = tHAND[a][origin]
  elseif origin_home == "Combat" and next(tCOMBAT[a][origin]) ~= nil then origin_card = tCOMBAT[a][origin]
  elseif origin_home == "Hero" and next(tHERO[a]) ~= nil then origin_card = tHERO[a]
  elseif origin_home == "Deck" and next(tDECK[a][origin]) ~= nil then origin_card = tDECK[a][origin]
  elseif origin_home == "Grave" and next(tGRAVE[a][origin]) ~= nil then origin_card = tGRAVE[a][origin]
  else end

  click.Card = origin_card
  click.Source = origin_home
  click.Player = a
  click.Index = origin
  table.insert(tCLICK, 1, deepcopy(click))

  if target > 0 then
    if target_home == "Hand" and next(tHAND[target_player][target]) ~= nil then target_card = tHAND[target_player][target]
    elseif target_home == "Combat" and next(tCOMBAT[target_player][target]) ~= nil then target_card = tCOMBAT[target_player][target]
    elseif target_home == "Hero" and next(tHERO[target_player]) ~= nil then target_card = tHERO[target_player]
    elseif target_home == "Deck" and next(tDECK[target_player][target]) ~= nil then target_card = tDECK[target_player][target]
    elseif target_home == "Grave" and next(tGRAVE[target_player][target]) ~= nil then target_card = tGRAVE[target_player][target]
    else end
    click = {}
    click.Card = target_card
    click.Source = target_home
    click.Player = target_player
    click.Index = target
    table.insert(tCLICK, 1, deepcopy(click))
  end

  if move == "Sacrifice" and next(origin_card) ~= nil then
    addpopup("P"..gTURN.." sacrifices "..origin_card.Name)
    sendgraveyard(origin_card, origin_home, a, origin)
    tBUDGET[a] = tBUDGET[a] + 1
    tPOINTS[a] = tPOINTS[a] + 1
    r = true
    gSACRIFICED = true
  elseif move == "Play to Combat" and next(origin_card) ~= nil then
    addpopup("P"..gTURN.." plays "..origin_card.Name.." to combat")
    sendcombat(origin_card, origin_home, a, origin)
    tPOINTS[a] = tPOINTS[a] - origin_card.Tier
    r = true
  elseif move == "Play action" and next(origin_card) ~= nil then
    gACTION = origin_card.Special
    if origin_card.SpTargets == 0 then
      if playaction(origin_card, a) == true then
        tPOINTS[a] = tPOINTS[a] - origin_card.Tier
        sendgraveyard(origin_card, origin_home, a, origin)
        r = true
      end
    elseif next(target_card) ~= nil then
      if playaction(origin_card, a) == true then
        tPOINTS[a] = tPOINTS[a] - origin_card.Tier
        dotarget()
        sendgraveyard(origin_card, origin_home, a, origin)
        r = true
      end
    else end
  elseif move == "Use" and next(origin_card) ~= nil then
    gACTION = origin_card.Special
    if origin_card.SpTargets == 0 then
      if playaction(origin_card, a) == true then
        tPOINTS[a] = tPOINTS[a] - origin_card.SpVal
        origin_card.SpecialUsed = true
        r = true
      end
    elseif next(target_card) ~= nil then
      if playaction(origin_card, a) == true then
        tPOINTS[a] = tPOINTS[a] - origin_card.SpVal
        dotarget()
        origin_card.SpecialUsed = true
        r = true
      end
    else end
  elseif move == "Attack" and next(origin_card) ~= nil and next(target_card) ~= nil then
    dobattle(origin_card, origin_home, a, origin, target_card, target_home, b, target)
    origin_card.Used = true
    r = true
  else end

  tCLICK = {}

  return r
end

function AIreport(move)
  local report = ""
  local mycard = nil
  local theircard = nil
  local a = gTURN  -- home player
  local b = move[6]  -- target player (can be the same)

  if move[3] == "Hand" then mycard = tHAND[a][move[2]]
  elseif move[3] == "Combat" then mycard = tCOMBAT[a][move[2]]
  elseif move[3] == "Hero" then mycard = tHERO[a]
  elseif move[3] == "Grave" then mycard = tGRAVE[a][move[2]]
  elseif move[3] == "Deck" then mycard = tDECK[a][move[2]]
  else mycard = nil end

  if move[4] > 0 then
    if move[5] == "Hand" then theircard = tHAND[b][move[4]]
    elseif move[5] == "Combat" then theircard = tCOMBAT[b][move[4]]
    elseif move[5] == "Hero" then theircard = tHERO[b]
    elseif move[5] == "Grave" then theircard = tGRAVE[b][move[4]]
    elseif move[5] == "Deck" then theircard = tDECK[b][move[4]]
    else theircard = nil end
  end

  if move[1] == "Play to Combat" then report = "March "..mycard.Name..": "..move[7]
  elseif move[1] == "Play action" and theircard ~= nil then report = "Play action ("..mycard.Special..") using "..mycard.Name.." on "..theircard.Name..": "..move[7]
  elseif move[1] == "Play action" and theircard == nil then report = "Play action ("..mycard.Special..") using "..mycard.Name..": "..move[7]
  elseif move[1] == "Attack" and theircard ~= nil then report = "Attack "..theircard.Name.." using "..mycard.Name..": "..move[7]
  elseif move[1] == "Sacrifice" then report = "Sacrifice "..mycard.Name..": "..move[7]
  elseif move[1] == "Use" and theircard ~= nil then report = "Use ability ("..mycard.Special..") using "..mycard.Name.." on "..theircard.Name..": "..move[7]
  elseif move[1] == "Use" and theircard == nil then report = "Use ability ("..mycard.Special..") using "..mycard.Name..": "..move[7]
  else report = "Not sure what this move was!?" end

  return report
end

function checkpoint()
  bDECK = deepcopy(tDECK)
  bHAND = deepcopy(tHAND)
  bCOMBAT = deepcopy(tCOMBAT)
  bGRAVE = deepcopy(tGRAVE)
  bHERO = deepcopy(tHERO)
  bBUDGET = deepcopy(tBUDGET)
  bPOINTS = deepcopy(tPOINTS)
  bSACRIFICED = gSACRIFICED
  bCEASEFIRE = gCEASEFIRE
  bACTION = gACTION
end

function restorecheckpoint()
  tDECK = deepcopy(bDECK)
  tHAND = deepcopy(bHAND)
  tCOMBAT = deepcopy(bCOMBAT)
  tGRAVE = deepcopy(bGRAVE)
  tHERO = deepcopy(bHERO)
  tBUDGET = deepcopy(bBUDGET)
  tPOINTS = deepcopy(bPOINTS)
  gSACRIFICED = bSACRIFICED
  gCEASEFIRE = bCEASEFIRE
  gACTION = bACTION
end

function harfway()
  hDECK = deepcopy(tDECK)
  hHAND = deepcopy(tHAND)
  hCOMBAT = deepcopy(tCOMBAT)
  hGRAVE = deepcopy(tGRAVE)
  hHERO = deepcopy(tHERO)
  hBUDGET = deepcopy(tBUDGET)
  hPOINTS = deepcopy(tPOINTS)
  hSACRIFICED = gSACRIFICED
  hCEASEFIRE = gCEASEFIRE
  hACTION = gACTION
end

function restoreharfway()
  tDECK = deepcopy(hDECK)
  tHAND = deepcopy(hHAND)
  tCOMBAT = deepcopy(hCOMBAT)
  tGRAVE = deepcopy(hGRAVE)
  tHERO = deepcopy(hHERO)
  tBUDGET = deepcopy(hBUDGET)
  tPOINTS = deepcopy(hPOINTS)
  gSACRIFICED = hSACRIFICED
  gCEASEFIRE = hCEASEFIRE
  gACTION = hACTION
end